# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Murakami_2013
^^^^^^^^^^^^^

Minerals from Murakami 2013 and references therein.

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class periclase(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 11.24e-6,
            "K_0": 161.0e9,
            "Kprime_0": 3.9,
            "G_0": 130.9e9,
            "Gprime_0": 1.92,
            "molar_mass": 0.0403,
            "n": 2,
            "Debye_0": 773.0,
            "grueneisen_0": 1.5,
            "q_0": 1.5,
            "eta_s_0": 2.3,
        }
        Mineral.__init__(self)


class wuestite(Mineral):
    """
    Murakami 2013 and references therein
    """

    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 12.06e-6,
            "K_0": 152.0e9,
            "Kprime_0": 4.9,
            "G_0": 47.0e9,
            "Gprime_0": 0.7,
            "molar_mass": 0.0718,
            "n": 2,
            "Debye_0": 455.0,
            "grueneisen_0": 1.28,
            "q_0": 1.5,
            "eta_s_0": 0.8,
        }

        Mineral.__init__(self)


class mg_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 24.45e-6,
            "K_0": 253.0e9,
            "Kprime_0": 4.1,
            "G_0": 172.9e9,
            "Gprime_0": 1.56,
            "molar_mass": 0.1000,
            "n": 5,
            "Debye_0": 1100.0,
            "grueneisen_0": 1.4,
            "q_0": 1.4,
            "eta_s_0": 2.6,
        }

        Mineral.__init__(self)


class fe_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 25.49e-6,
            "K_0": 281.0e9,
            "Kprime_0": 4.1,
            "G_0": 138.0e9,
            "Gprime_0": 1.7,
            "molar_mass": 0.1319,
            "n": 5,
            "Debye_0": 841.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
            "eta_s_0": 2.1,
        }

        Mineral.__init__(self)


mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import warnings
from scipy.linalg import logm


def check_eos_consistency(
    m,
    P=1.0e9,
    T=300.0,
    tol=1.0e-4,
    verbose=False,
    including_shear_properties=True,
    equilibration_function=None,
):
    """
    Checks that numerical derivatives of the Gibbs energy of a mineral
    under given conditions are equal to those provided
    analytically by the equation of state.

    :param m: The mineral for which the equation of state
        is to be checked for consistency.
    :type m: :class:`burnman.Mineral`

    :param P: The pressure at which to check consistency.
    :type P: float

    :param T: The temperature at which to check consistency.
    :type T: float

    :param tol: The fractional tolerance for each of the checks.
    :type tol: float

    :param verbose: Decide whether to print information about each check.
    :type verbose: bool

    :param including_shear_properties: Decide whether to check shear information,
        which is pointless for liquids and equations of state
        without shear modulus parameterizations.
    :type including_shear_properties: bool

    :param equilibration_function: Function to internally equilibrate object.
        Called after every set_state.
        Takes the mineral object as its only argument.
    :type equilibration_function: function

    :returns: Boolean stating whether all checks have passed.
    :rtype: bool
    """
    if equilibration_function is None:

        def equilibration_function(mineral):
            pass

    dT = 1.0
    dP = 1000.0

    m.set_state(P, T)
    equilibration_function(m)
    G0 = m.gibbs
    S0 = m.S
    V0 = m.V

    expr = ["G = F + PV", "G = H - TS", "G = E - TS + PV"]
    eq = [
        [m.gibbs, (m.helmholtz + P * m.V)],
        [m.gibbs, (m.H - T * m.S)],
        [m.gibbs, (m.molar_internal_energy - T * m.S + P * m.V)],
    ]

    m.set_state(P, T + dT)
    equilibration_function(m)
    G1 = m.gibbs
    S1 = m.S
    V1 = m.V

    m.set_state(P + dP, T)
    equilibration_function(m)
    G2 = m.gibbs
    V2 = m.V

    # T derivatives
    m.set_state(P, T + 0.5 * dT)
    equilibration_function(m)
    expr.extend(["S = -dG/dT", "alpha = 1/V dV/dT", "C_p = T dS/dT"])
    eq.extend(
        [
            [m.S, -(G1 - G0) / dT],
            [m.alpha, (V1 - V0) / dT / m.V],
            [m.molar_heat_capacity_p, (T + 0.5 * dT) * (S1 - S0) / dT],
        ]
    )

    # P derivatives
    m.set_state(P + 0.5 * dP, T)
    equilibration_function(m)
    expr.extend(["V = dG/dP", "K_T = -V dP/dV"])
    eq.extend(
        [
            [m.V, (G2 - G0) / dP],
            [m.isothermal_bulk_modulus_reuss, -0.5 * (V2 + V0) * dP / (V2 - V0)],
        ]
    )

    expr.extend(
        ["C_v = Cp - alpha^2*K_T*V*T", "K_S = K_T*Cp/Cv", "gr = alpha*K_T*V/Cv"]
    )
    eq.extend(
        [
            [
                m.molar_heat_capacity_v,
                m.molar_heat_capacity_p
                - m.alpha * m.alpha * m.isothermal_bulk_modulus_reuss * m.V * T,
            ],
            [
                m.isentropic_bulk_modulus_reuss,
                m.isothermal_bulk_modulus_reuss
                * m.molar_heat_capacity_p
                / m.molar_heat_capacity_v,
            ],
            [
                m.gr,
                m.alpha
                * m.isothermal_bulk_modulus_reuss
                * m.V
                / m.molar_heat_capacity_v,
            ],
        ]
    )

    expr.append("Vphi = np.sqrt(K_S/rho)")
    eq.append([m.bulk_sound_velocity, np.sqrt(m.isentropic_bulk_modulus_reuss / m.rho)])

    if including_shear_properties:
        expr.extend(["Vp = np.sqrt((K_S + 4G/3)/rho)", "Vs = np.sqrt(G_S/rho)"])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eq.extend(
                [
                    [
                        m.p_wave_velocity,
                        np.sqrt(
                            (m.isentropic_bulk_modulus_reuss + 4.0 * m.G / 3.0) / m.rho
                        ),
                    ],
                    [m.shear_wave_velocity, np.sqrt(m.G / m.rho)],
                ]
            )
            if len(w) == 1:
                print(w[0].message)
                print(
                    "\nYou can suppress this message by setting the "
                    "parameter\nincluding_shear_properties to False "
                    "when calling check_eos_consistency.\n"
                )
        note = ""
    else:
        note = " (not including shear properties)"

    consistencies = [
        np.abs(e[0] - e[1]) < np.abs(tol * e[1]) + np.finfo("float").eps for e in eq
    ]
    eos_is_consistent = np.all(consistencies)

    if verbose:
        print("Checking EoS consistency for {0:s}{1}".format(m.to_string(), note))
        print("Expressions within tolerance of {0:2f}".format(tol))
        for i, c in enumerate(consistencies):
            print("{0:10s} : {1:5s}".format(expr[i], str(c)))
            if not c:
                print(eq[i])
        if eos_is_consistent:
            print(
                "All EoS consistency constraints satisfied for {0:s}".format(
                    m.to_string()
                )
            )
        else:
            print(
                "Not satisfied all EoS consistency constraints for {0:s}".format(
                    m.to_string()
                )
            )

    return eos_is_consistent


def check_anisotropic_eos_consistency(
    m, P=1.0e9, T=300.0, tol=1.0e-4, verbose=False, equilibration_function=None
):
    """
    Checks that numerical derivatives of the Gibbs energy of an anisotropic mineral
    under given conditions are equal to those provided
    analytically by the equation of state.

    :param m: The anisotropic mineral for which the equation of state
        is to be checked for consistency.
    :type m: :class:`burnman.AnisotropicMineral`

    :param P: The pressure at which to check consistency.
    :type P: float

    :param T: The temperature at which to check consistency.
    :type T: float

    :param tol: The fractional tolerance for each of the checks.
    :type tol: float

    :param verbose: Decide whether to print information about each check.
    :type verbose: bool

    :param equilibration_function: Function to internally equilibrate object.
        Called after every set_state.
        Takes the mineral object as its only argument.
    :type equilibration_function: function

    :returns: Boolean stating whether all checks have passed.
    :rtype: bool
    """
    if equilibration_function is None:

        def equilibration_function(mineral):
            pass

    dT = 1.0
    dP = 1000.0

    m.set_state(P, T)
    equilibration_function(m)
    G0 = m.gibbs
    S0 = m.S
    V0 = m.V

    expr = ["G = F + PV", "G = H - TS", "G = E - TS + PV"]
    eq = [
        [m.gibbs, (m.helmholtz + P * m.V)],
        [m.gibbs, (m.H - T * m.S)],
        [m.gibbs, (m.molar_internal_energy - T * m.S + P * m.V)],
    ]

    m.set_state(P, T + dT)
    equilibration_function(m)
    G1 = m.gibbs
    S1 = m.S
    V1 = m.V

    m.set_state(P + dP, T)
    equilibration_function(m)
    G2 = m.gibbs
    V2 = m.V

    # T derivatives
    m.set_state(P, T + 0.5 * dT)
    equilibration_function(m)
    expr.extend(["S = -dG/dT", "alpha = 1/V dV/dT", "C_p = T dS/dT"])
    eq.extend(
        [
            [m.S, -(G1 - G0) / dT],
            [m.alpha, (V1 - V0) / dT / m.V],
            [m.molar_heat_capacity_p, (T + 0.5 * dT) * (S1 - S0) / dT],
        ]
    )

    # P derivatives
    m.set_state(P + 0.5 * dP, T)
    equilibration_function(m)
    expr.extend(["V = dG/dP", "K_T = -V dP/dV"])
    eq.extend(
        [
            [m.V, (G2 - G0) / dP],
            [m.isothermal_bulk_modulus_reuss, -0.5 * (V2 + V0) * dP / (V2 - V0)],
        ]
    )

    expr.extend(["C_v = Cp - alpha^2*K_T*V*T", "K_S = K_T*Cp/Cv"])
    eq.extend(
        [
            [
                m.molar_heat_capacity_v,
                m.molar_heat_capacity_p
                - m.alpha * m.alpha * m.isothermal_bulk_modulus_reuss * m.V * T,
            ],
            [
                m.isentropic_bulk_modulus_reuss,
                m.isothermal_bulk_modulus_reuss
                * m.molar_heat_capacity_p
                / m.molar_heat_capacity_v,
            ],
        ]
    )

    # Third derivative
    m.set_state(P + 0.5 * dP, T)
    equilibration_function(m)
    b0 = m.isothermal_compressibility_tensor
    F0 = m.deformation_gradient_tensor

    m.set_state(P + 0.5 * dP, T + dT)
    equilibration_function(m)
    b1 = m.isothermal_compressibility_tensor
    F1 = m.deformation_gradient_tensor

    m.set_state(P, T + 0.5 * dT)
    equilibration_function(m)
    a0 = m.thermal_expansivity_tensor
    F2 = m.deformation_gradient_tensor

    m.set_state(P + dP, T + 0.5 * dT)
    equilibration_function(m)
    a1 = m.thermal_expansivity_tensor
    F3 = m.deformation_gradient_tensor

    m.set_state(P + 0.5 * dP, T + 0.5 * dT)
    equilibration_function(m)

    beta0 = -(logm(F3) - logm(F2)) / dP
    alpha0 = (logm(F1) - logm(F0)) / dT

    Q = m.deformed_coordinate_frame
    beta1 = m.isothermal_compressibility_tensor
    alpha1 = m.thermal_expansivity_tensor

    beta1 = np.einsum("mi, nj, ij->mn", Q, Q, beta1)
    alpha1 = np.einsum("mi, nj, ij->mn", Q, Q, alpha1)

    if m.orthotropic:
        expr.extend(
            [f"SI = -d(lnm(F))/dP ({i}{j})" for i in range(3) for j in range(i, 3)]
        )
        expr.extend(
            [f"alpha = d(lnm(F))/dT ({i}{j})" for i in range(3) for j in range(i, 3)]
        )
    else:
        expr.extend(
            [
                f"SI = -0.5(dF/dP*F^-1 + (dF/dP*F^-1)^T) ({i}{j})"
                for i in range(3)
                for j in range(i, 3)
            ]
        )
        expr.extend(
            [
                f"alpha = 0.5(dF/dT*F^-1 + (dF/dT*F^-1)^T) ({i}{j})"
                for i in range(3)
                for j in range(i, 3)
            ]
        )
        invF = np.linalg.inv(m.deformation_gradient_tensor)
        dFdP = (F3 - F2) / dP
        LP = np.einsum("ij,kj->ik", dFdP, invF)
        beta0 = -0.5 * (LP + LP.T)
        dFdT = (F1 - F0) / dT
        LT = np.einsum("ij,kj->ik", dFdT, invF)
        alpha0 = 0.5 * (LT + LT.T)

    eq.extend([[beta0[i, j], beta1[i, j]] for i in range(3) for j in range(i, 3)])
    eq.extend([[alpha0[i, j], alpha1[i, j]] for i in range(3) for j in range(i, 3)])

    expr.extend(
        [f"d(alpha)/dP = -d(beta_T)/dT ({i}{j})" for i in range(3) for j in range(i, 3)]
    )
    eq.extend(
        [
            [(a1[i, j] - a0[i, j]) / dP, -(b1[i, j] - b0[i, j]) / dT]
            for i in range(3)
            for j in range(i, 3)
        ]
    )

    # Consistent Phi
    expr.extend(["dPsidf_Voigt[:3,:3] == 1"])
    eq.extend(
        [
            [
                np.sum(m.isothermal_compliance_tensor[:3, :3]),
                m.isothermal_compressibility_reuss,
            ]
        ]
    )

    # Consistent inverses
    expr.extend([f"S_T = inv(C_T) ({i}{j})" for i in range(6) for j in range(i, 6)])
    S_T = m.isothermal_compliance_tensor
    S_T2 = np.linalg.inv(m.isothermal_stiffness_tensor)
    eq.extend([[S_T[i, j], S_T2[i, j]] for i in range(6) for j in range(i, 6)])

    expr.extend([f"S_N = inv(C_N) ({i}{j})" for i in range(6) for j in range(i, 6)])
    S_N = m.isentropic_compliance_tensor
    S_N2 = np.linalg.inv(m.isentropic_stiffness_tensor)
    eq.extend([[S_N[i, j], S_N2[i, j]] for i in range(6) for j in range(i, 6)])

    # Consistent isotropic and anisotropic properties
    expr.extend(
        [
            "V = det(M)",
            "alpha_v = tr(alpha)",
            "beta_RT = sum(I S_T I)",
            "beta_RS = sum(I S_S I)",
        ]
    )
    eq.extend(
        [
            [m.V, np.linalg.det(m.cell_vectors)],
            [m.alpha, np.trace(m.thermal_expansivity_tensor)],
            [
                m.isothermal_compressibility_reuss,
                np.sum(m.isothermal_compliance_tensor[:3, :3]),
            ],
            [
                m.isentropic_compressibility_reuss,
                np.sum(m.isentropic_compliance_tensor[:3, :3]),
            ],
        ]
    )

    expr.append("Vphi = np.sqrt(K_S/rho)")
    eq.append([m.bulk_sound_velocity, np.sqrt(m.isentropic_bulk_modulus_reuss / m.rho)])

    consistencies = [
        np.abs(e[0] - e[1]) < np.abs(tol * e[1]) + np.finfo("float").eps for e in eq
    ]
    eos_is_consistent = np.all(consistencies)

    if verbose:
        print("Checking EoS consistency for {0:s}".format(m.to_string()))
        print("Expressions within tolerance of {0:2f}".format(tol))
        for i, c in enumerate(consistencies):
            print("{0:10s} : {1:5s}".format(expr[i], str(c)))
            if not c:
                print(eq[i])
        if eos_is_consistent:
            print(
                "All EoS consistency constraints satisfied for {0:s}".format(
                    m.to_string()
                )
            )
        else:
            print(
                "Not satisfied all EoS consistency constraints for {0:s}".format(
                    m.to_string()
                )
            )

    return eos_is_consistent



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np

from . import nonlinear_fitting
from ..utils.misc import flatten
from ..utils.math import unit_normalize
from .nonlinear_fitting import nonlinear_least_squares_fit


class MineralFit(object):
    """
    Class for fitting mineral parameters to experimental data.
    Instances of this class are passed to
    :func:`burnman.nonlinear_least_squares_fit`.

    For attributes added to this model when fitting is done,
    please see the documentation for that function.
    """

    def __init__(
        self,
        mineral,
        data,
        data_covariances,
        flags,
        fit_params,
        mle_tolerances,
        delta_params=None,
        bounds=None,
    ):
        self.m = mineral
        self.data = data
        self.data_covariances = data_covariances
        self.flags = flags
        self.fit_params = fit_params
        self.mle_tolerances = mle_tolerances
        if delta_params is None:
            self.delta_params = self.get_params() * 1.0e-5 + 1.0e-10
        else:
            self.delta_params = delta_params
        self.bounds = bounds

    def set_params(self, param_values):
        i = 0

        if self.bounds is not None:
            param_values = np.clip(param_values, self.bounds[:, 0], self.bounds[:, 1])

        for param in self.fit_params:
            if isinstance(self.m.params[param], float):
                self.m.params[param] = param_values[i]
                i += 1
            else:
                for j in range(len(self.m.params[param])):
                    self.m.params[param][j] = param_values[i]
                    i += 1

    def get_params(self):
        params = []
        for i, param in enumerate(self.fit_params):
            params.append(self.m.params[param])
        return np.array(flatten([self.m.params[prm] for prm in self.fit_params]))

    def function(self, x, flag):
        P, T, p = x
        self.m.set_state(P, T)
        return np.array([P, T, getattr(self.m, flag)])

    def normal(self, x, flag):
        P, T, p = x

        if flag == "V":
            self.m.set_state(P, T)
            dPdp = -self.m.isothermal_bulk_modulus_reuss / self.m.V
            dpdT = self.m.alpha * self.m.V
        elif flag == "H":
            self.m.set_state(P, T)
            dPdp = 1.0 / ((1.0 - T * self.m.alpha) * self.m.V)
            dpdT = self.m.molar_heat_capacity_p
        elif flag == "S":
            self.m.set_state(P, T)
            dPdp = -1.0 / (self.m.alpha * self.m.V)
            dpdT = self.m.molar_heat_capacity_p / T
        elif flag == "gibbs":
            self.m.set_state(P, T)
            dPdp = 1.0 / self.m.V
            dpdT = -self.m.S
        else:
            dP = 1.0e5
            dT = 1.0
            dPdp = (2.0 * dP) / (
                self.function([P + dP, T, 0.0], flag)[2]
                - self.function([P - dP, T, 0.0], flag)[2]
            )
            dpdT = (
                self.function([P, T + dT, 0.0], flag)[2]
                - self.function([P, T - dT, 0.0], flag)[2]
            ) / (2.0 * dT)
        dPdT = -dPdp * dpdT
        n = np.array([-1.0, dPdT, dPdp])
        return unit_normalize(n)


def fit_PTp_data(
    mineral,
    fit_params,
    flags,
    data,
    data_covariances=[],
    mle_tolerances=[],
    param_tolerance=1.0e-5,
    delta_params=None,
    bounds=None,
    max_lm_iterations=50,
    verbose=True,
):
    """
    Given a mineral of any type, a list of fit parameters
    and a set of P-T-property points and (optional) uncertainties,
    this function returns a list of optimized parameters
    and their associated covariances, fitted using the
    scipy.optimize.curve_fit routine.

    :param mineral: Mineral for which the parameters should be optimized.
    :type mineral: :class:`burnman.Mineral`

    :param fit_params: List of dictionary keys contained in mineral.params
        corresponding to the variables to be optimized
        during fitting. Initial guesses are taken from the existing
        values for the parameters
    :type fit_params: list of str

    :param flags: Attribute names for the property to be fit for the whole
        dataset or each datum individually (e.g. 'V')
    :type flags: string or list of strings

    :param data: Observed X-P-T-property values
    :type data: 2D numpy.array

    :param data_covariances: X-P-T-property covariances (optional)
        If not given, all covariance matrices are chosen
        such that all data points have equal weight,
        with all error in the pressure.
    :type data_covariances: 3D numpy.array

    :param mle_tolerances: Tolerances for termination of the
        maximum likelihood iterations (optional).
    :type mle_tolerances: numpy.array

    :param param_tolerance: Fractional tolerance for termination
        of the nonlinear optimization (optional).
    :type param_tolerance: float

    :param delta_params: Initial values for the change in parameters (optional).
    :type delta_params: numpy.array

    :param bounds: Minimum and maximum bounds for the parameters (optional).
        The shape must be (n_parameters, 2).
    :type bounds: 2D numpy.array

    :param max_lm_iterations: Maximum number of Levenberg-Marquardt iterations.
    :type max_lm_iterations: int

    :param verbose: Whether to print detailed information about the
        optimization to screen.
    :type verbose: bool

    :returns: Model with optimized parameters.
    :rtype: :class:`burnman.optimize.eos_fitting.MineralFit`
    """

    # If only one property flag is given, assume it applies to all data
    if type(flags) is str:
        flags = np.array([flags] * len(data[:, 0]))

    if len(flags) != len(data):
        raise Exception(
            f"The number of flags (n = {len(flags)}) must be equal "
            f"to the number of data (n = {len(data)})."
        )

    # Apply mle tolerances if they dont exist
    if len(mle_tolerances) == 0:
        mineral.set_state(1.0e5, 300.0)
        mle_tolerance_factor = 1.0e-5
        mle_tolerances = np.empty(len(flags))
        for i, flag in enumerate(flags):
            if flag in ["gibbs", "enthalpy", "H", "helmholtz"]:
                mle_tolerances[i] = 1.0  # 1 J
            else:
                mle_tolerances[i] = mle_tolerance_factor * getattr(mineral, flag)

    # If covariance matrix is not given, apply unit weighting to all pressures
    # (with zero errors on T and p)
    covariances_defined = True
    if len(data_covariances) == 0:
        covariances_defined = False
        data_covariances = np.zeros((len(data[:, 0]), len(data[0]), len(data[0])))
        for i in range(len(data_covariances)):
            data_covariances[i][0][0] = 1.0

    model = MineralFit(
        mineral=mineral,
        data=data,
        data_covariances=data_covariances,
        flags=flags,
        fit_params=fit_params,
        delta_params=delta_params,
        mle_tolerances=mle_tolerances,
        bounds=bounds,
    )

    nonlinear_least_squares_fit(
        model,
        max_lm_iterations=max_lm_iterations,
        param_tolerance=param_tolerance,
        verbose=verbose,
    )

    if verbose is True and covariances_defined is True:
        confidence_interval = 0.9
        d = nonlinear_fitting.extreme_values(
            model.weighted_residuals, confidence_interval
        )
        confidence_bound, indices, probabilities = d
        if indices != []:
            print(
                "The function nonlinear_fitting.extreme_values"
                "(model.weighted_residuals, confidence_interval) "
                f"has determined that there are {len(indices):d} data points"
                " which have residuals which are not expected at the "
                f"{confidence_interval*100.:.1f}% confidence level "
                f"(> {confidence_bound:.1f} s.d. away from the model fit).\n"
                "Their indices and the probabilities of finding "
                "such extreme values are:"
            )
            for i, idx in enumerate(indices):
                print(
                    f"[{idx:d}]: {probabilities[i]:.4f} "
                    f"({np.abs(model.weighted_residuals[idx]):.1f} s.d. "
                    "from the model)"
                )
            print(
                "You might consider removing them from your fit, "
                "or increasing the uncertainties in their "
                "measured values.\n"
            )

    return model


def fit_PTV_data(
    mineral,
    fit_params,
    data,
    data_covariances=[],
    delta_params=None,
    bounds=None,
    param_tolerance=1.0e-5,
    max_lm_iterations=50,
    verbose=True,
):
    """
    A simple alias for the fit_PTp_data for when all the data is volume data
    """

    return fit_PTp_data(
        mineral=mineral,
        flags="V",
        data=data,
        data_covariances=data_covariances,
        fit_params=fit_params,
        param_tolerance=param_tolerance,
        delta_params=delta_params,
        bounds=bounds,
        max_lm_iterations=max_lm_iterations,
        verbose=verbose,
    )


class SolutionFit(object):
    """
    Class for fitting mineral parameters to experimental data.
    Instances of this class are passed to
    :func:`burnman.nonlinear_least_squares_fit`.

    For attributes added to this model when fitting is done,
    please see the documentation for that function.
    """

    def __init__(
        self,
        solution,
        data,
        data_covariances,
        flags,
        fit_params,
        mle_tolerances,
        delta_params=None,
        bounds=None,
    ):
        self.m = solution
        self.data = data
        self.data_covariances = data_covariances
        self.flags = flags
        self.fit_params = fit_params
        self.fit_params_strings = []
        for p in fit_params:
            if isinstance(p, list):
                csv_list_mbrs = ",".join([str(i) for i in p[1:]])
                self.fit_params_strings.append(f"{p[0]} ({csv_list_mbrs})")
            else:
                self.fit_params_strings.append(p)

        self.mle_tolerances = mle_tolerances
        if delta_params is None:
            self.delta_params = self.get_params() * 1.0e-5 + 1.0e-10
        else:
            self.delta_params = delta_params
        self.bounds = bounds

    def set_params(self, param_values):
        # fit_params is a list of lists
        # if the list has length 2, the first item should be an integer
        # indicating the endmember number in the solution
        # if the list has length 3, the first two items should be endmember
        # numbers, and the third should be the interaction parameter type
        # (E, S or V).
        i = 0

        if self.bounds is not None:
            param_values = np.clip(param_values, self.bounds[:, 0], self.bounds[:, 1])

        for param in self.fit_params:
            value = param_values[i]
            if len(param) == 2:
                key, imbr = param
                if isinstance(self.m.endmembers[imbr][0].params[key], float):
                    self.m.endmembers[imbr][0].params[key] = value
                    i += 1
                else:
                    n_values = len(self.m.endmembers[imbr][0].params[key])
                    for j in range(n_values):
                        self.m.endmembers[imbr][0].params[key][j] = value
                        i += 1
            elif len(param) == 3:
                key, imbr, jmbr = param
                ai = self.m.solution_model.alphas[imbr]
                aj = self.m.solution_model.alphas[jmbr]
                if key == "E":
                    self.m.solution_model.We[imbr, jmbr] = 2.0 * value / (ai * aj)
                if key == "S":
                    self.m.solution_model.Ws[imbr, jmbr] = 2.0 * value / (ai * aj)
                if key == "V":
                    self.m.solution_model.Wv[imbr, jmbr] = 2.0 * value / (ai * aj)

                i += 1
            else:
                raise Exception("param length must be two or three")

    def get_params(self):
        params = []
        for param in self.fit_params:
            if len(param) == 2:
                key, imbr = param
                value = self.m.endmembers[imbr][0].params[key]
                if isinstance(value, float):
                    params.append(value)
                else:
                    params.extend(list(value))

            elif len(param) == 3:
                key, imbr, jmbr = param
                ai = self.m.solution_model.alphas[imbr]
                aj = self.m.solution_model.alphas[jmbr]
                if key == "E":
                    params.append(
                        self.m.solution_model.We[imbr, jmbr] * (ai * aj) / 2.0
                    )
                if key == "S":
                    params.append(
                        self.m.solution_model.Ws[imbr, jmbr] * (ai * aj) / 2.0
                    )
                if key == "V":
                    params.append(
                        self.m.solution_model.Wv[imbr, jmbr] * (ai * aj) / 2.0
                    )
            else:
                raise Exception("param length must be two or three")
        return np.array(params)

    def function(self, x, flag):
        self.m.set_composition(x[: self.m.n_endmembers])
        P, T, p = x[self.m.n_endmembers :]
        self.m.set_state(P, T)

        f = np.copy(x)
        f[-1] = getattr(self.m, flag)
        return f

    def normal(self, x, flag):
        self.m.set_composition(x[: self.m.n_endmembers])
        P, T, p = x[self.m.n_endmembers :]

        if flag == "V":
            self.m.set_state(P, T)
            dPdp = -self.m.isothermal_bulk_modulus_reuss / self.m.V
            dpdT = self.m.alpha * self.m.V
        elif flag == "H":
            self.m.set_state(P, T)
            dPdp = 1.0 / ((1.0 - T * self.m.alpha) * self.m.V)
            dpdT = self.m.molar_heat_capacity_p
        elif flag == "S":
            self.m.set_state(P, T)
            dPdp = -1.0 / (self.m.alpha * self.m.V)
            dpdT = self.m.molar_heat_capacity_p / T
        elif flag == "gibbs":
            self.m.set_state(P, T)
            dPdp = 1.0 / self.m.V
            dpdT = -self.m.S
        else:
            dP = 1.0e5
            dT = 1.0
            xP0 = np.copy(x)
            xP1 = np.copy(x)
            xT0 = np.copy(x)
            xT1 = np.copy(x)
            xP0[-3] = xP1[-3] - dP
            xP1[-3] = xP1[-3] + dP
            xT0[-2] = xP1[-2] - dT
            xT1[-2] = xP1[-2] + dT

            dPdp = (2.0 * dP) / (
                self.function(xP1, flag)[2] - self.function(xP0, flag)[2]
            )
            dpdT = (self.function(xT1, flag)[2] - self.function(xT0, flag)[2]) / (
                2.0 * dT
            )
        dPdT = -dPdp * dpdT
        n = np.zeros(len(x))
        n[-3:] = np.array([-1.0, dPdT, dPdp])
        return unit_normalize(n)


def fit_XPTp_data(
    solution,
    fit_params,
    flags,
    data,
    data_covariances=[],
    mle_tolerances=[],
    param_tolerance=1.0e-5,
    delta_params=None,
    bounds=None,
    max_lm_iterations=50,
    verbose=True,
):
    """
    Given a symmetric solution, a list of fit parameters
    and a set of P-T-property points and (optional) uncertainties,
    this function returns a list of optimized parameters
    and their associated covariances, fitted using the
    scipy.optimize.curve_fit routine.

    :param solution: Solution for which the parameters should be optimized.
    :type solution: :class:`burnman.Solution`

    :param fit_params: Variables to be optimized
        during fitting. Each list is either of length two or three.
        The first item of length-2 lists should be a
        dictionary key contained in one of the endmember
        mineral.params, and the second item should be the index of
        the endmember in the solution (indexing starts from 0).
        The first item of length-3 lists should be one of 'E', 'S' or
        'V' (the excess energies, entropies or volumes in each binary).
        The second two items should be the indices of the pair of
        endmembers bounding the binary, in ascending order
        (indexing starts from 0). Initial guesses are taken from the existing
        values for the parameters.
    :type fit_params: list of lists

    :param flags: Attribute names for the property to be fit for the whole
        dataset or each datum individually (e.g. 'V')
    :type flags: string or list of strings

    :param data: Observed X-P-T-property values
    :type data: 2D numpy.array

    :param data_covariances: X-P-T-property covariances (optional).
        If not given, all covariance matrices are chosen
        such that all data points have equal weight,
        with all error in the pressure.
    :type data_covariances: 3D numpy.array

    :param mle_tolerances: Tolerances for termination of the
        maximum likelihood iterations (optional).
    :type mle_tolerances: numpy.array

    :param param_tolerance: Fractional tolerance for termination
        of the nonlinear optimization (optional).
    :type param_tolerance: float

    :param delta_params: Initial values for the change in parameters (optional).
    :type delta_params: numpy.array

    :param bounds: Minimum and maximum bounds for the parameters (optional).
        The shape must be (n_parameters, 2).
    :type bounds: 2D numpy.array

    :param max_lm_iterations: Maximum number of Levenberg-Marquardt iterations.
    :type max_lm_iterations: int

    :param verbose: Whether to print detailed information about the
        optimization to screen.
    :type verbose: bool

    :returns: Model with optimized parameters.
    :rtype: :class:`burnman.optimize.eos_fitting.SolutionFit`
    """
    # If only one property flag is given, assume it applies to all data
    if type(flags) is str:
        flags = np.array([flags] * len(data[:, 0]))

    if len(flags) != len(data):
        raise Exception(
            f"The number of flags (n = {len(flags)}) must be equal "
            f"to the number of data (n = {len(data)})."
        )

    # Apply mle tolerances if they dont exist
    if len(mle_tolerances) == 0:
        solution.set_state(1.0e5, 300.0)
        mle_tolerance_factor = 1.0e-5
        mle_tolerances = np.empty(len(flags))
        for i, flag in enumerate(flags):
            if flag in ["gibbs", "enthalpy", "H", "helmholtz"]:
                mle_tolerances[i] = 1.0  # 1 J
            else:
                mle_tolerances[i] = mle_tolerance_factor * getattr(solution, flag)

    # If covariance matrix is not given, apply unit weighting to all pressures
    # (with zero errors on T and property)
    covariances_defined = True
    if len(data_covariances) == 0:
        covariances_defined = False
        nX = solution.n_endmembers
        data_covariances = np.zeros((len(data[:, 0]), len(data[0]), len(data[0])))
        for i in range(len(data_covariances)):
            data_covariances[i][nX][nX] = 1.0

    model = SolutionFit(
        solution=solution,
        data=data,
        data_covariances=data_covariances,
        flags=flags,
        fit_params=fit_params,
        mle_tolerances=mle_tolerances,
        delta_params=delta_params,
        bounds=bounds,
    )

    nonlinear_least_squares_fit(
        model,
        max_lm_iterations=max_lm_iterations,
        param_tolerance=param_tolerance,
        verbose=verbose,
    )

    if verbose is True and covariances_defined is True:
        confidence_interval = 0.9
        v_extreme = nonlinear_fitting.extreme_values(
            model.weighted_residuals, confidence_interval
        )
        confidence_bound, indices, probabilities = v_extreme
        if indices != []:
            print(
                "The function nonlinear_fitting.extreme_values"
                "(model.weighted_residuals, confidence_interval) "
                f"has determined that there are {len(indices):d} "
                "data points which have residuals which are not "
                f"expected at the {confidence_interval*100.:.1f}% "
                "confidence level "
                f"(> {confidence_bound:.1f} s.d. away from the model fit).\n"
                "Their indices and the probabilities of "
                "finding such extreme values are:"
            )
            for i, idx in enumerate(indices):
                print(
                    f"[{idx:d}]: {probabilities[i]:.4f} "
                    f"({np.abs(model.weighted_residuals[idx]):.1f} s.d. "
                    "from the model)"
                )
            print(
                "You might consider removing them from your fit, "
                "or increasing the uncertainties in "
                "their measured values.\n"
            )

    return model



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.
import numpy as np
from scipy.linalg import expm, logm
from numpy.linalg import cond
from .mineral import Mineral
from .material import Material, material_property
from .anisotropy import AnisotropicMaterial
from ..utils.misc import copy_documentation
from ..utils.unitcell import cell_parameters_to_vectors
from ..utils.unitcell import cell_vectors_to_parameters
from ..utils.anisotropy import (
    voigt_notation_to_compliance_tensor,
    voigt_notation_to_stiffness_tensor,
    contract_compliances,
)


def convert_f_Pth_to_f_T_derivatives(
    dPsidf_Pth_Voigt, dPsidPth_f_Voigt, aK_T, dPthdf_T
):
    """
    Convenience function converting Psi and its derivatives with respect to
    f and Pth into derivatives of f and T using the chain rule.

    :param dPsidf_Pth_Voigt: The first derivative of the anisotropic tensor
        Psi with respect to f at constant Pth (Voigt form).
    :type dPsiIdf_Pth: numpy array (6x6)
    :param dPsidPth_f_Voigt: The first derivative of the anisotropic tensor
        Psi with respect to Pth at constant f (Voigt form).
    :type dPsiIdPth_f: numpy array (6x6)
    :param aK_T: The volumetric thermal expansivity multiplied by
        the Reuss isothermal bulk modulus.
    :type aK_T: float
    :param dPthdf: The change in thermal pressure with respect to volume
        at constant temperature.
    :type dPthdf: float

    :returns: The first derivative of Psi with respect to f in Voigt form,
        and the first derivatives of PsiI with respect to f and T.
    :rtype: Tuple of three objects of type numpy.array (2D)
    """
    # The following manipulation uses the chain rule to convert
    # from Psi(f, Pth) and derivatives to Psi(f, T) and derivatives
    dPsidf_full = voigt_notation_to_compliance_tensor(dPsidf_Pth_Voigt)
    dPsidPth_full = voigt_notation_to_compliance_tensor(dPsidPth_f_Voigt)

    dPsiIdf_Pth = np.einsum("ijkl, kl", dPsidf_full, np.eye(3))
    dPsiIdPth_f = np.einsum("ijkl, kl", dPsidPth_full, np.eye(3))

    dPsidf_T_Voigt = dPsidf_Pth_Voigt + dPsidPth_f_Voigt * dPthdf_T
    dPsiIdf_T = dPsiIdf_Pth + dPsiIdPth_f * dPthdf_T
    dPsiIdT_f = aK_T * dPsiIdPth_f
    return (dPsidf_T_Voigt, dPsiIdf_T, dPsiIdT_f)


def deformation_gradient_alpha_and_compliance(
    alpha_V, beta_TR, PsiI, dPsidf_T_Voigt, dPsiIdf_T, dPsiIdT_f
):
    """
    Convenience function converting Psi and its derivatives with respect to
    f and T into the deformation gradient, thermal expansivity and
    isothermal compliance tensors. This function is appropriate for
    both orthotropic and non-orthotropic materials.

    :param alpha_V: The volumetric thermal expansivity.
    :type alpha_V: float
    :param beta_TR: The Reuss isothermal compressibility.
    :type beta_TR: float
    :param psiI: The anisotropic tensor Psi matrix multiplied with
        the identity matrix.
    :type psiI: numpy array (3x3)
    :param dPsidf_T_Voigt: Voigt-form matrix of the first
        derivative of the anisotropic tensor Psi with respect to f
        at constant T.
    :type dPsidf_T_Voigt: numpy array (6x6)
    :param dPsiIdf_T: The first derivative of PsiI
        with respect to f at constant T.
    :type dPsiIdf_T: numpy array (3x3)
    :param dPsiIdT_f: The first derivative of PsiI
        with respect to T at constant f.
    :type dPsiIdT_f: numpy array (3x3)

    :returns: The deformation gradient tensor (3x3),
        derivative with respect to lnV (3x3),
        the thermal expansivity tensor (3x3),
        and the unrotated isothermal compliance tensor in Voigt form (6x6).
    :rtype: Tuple of four objects of type numpy.array (2D)
    """
    # Numerical derivatives with respect to f and T
    df = 1.0e-7
    F0 = expm(PsiI - dPsiIdf_T * df / 2.0)
    F1 = expm(PsiI + dPsiIdf_T * df / 2.0)
    dFdf_T = (F1 - F0) / df

    dT = 0.1
    F0 = expm(PsiI - dPsiIdT_f * dT / 2.0)
    F1 = expm(PsiI + dPsiIdT_f * dT / 2.0)
    dFdT_f = (F1 - F0) / dT

    # Convert to pressure and temperature derivatives
    dFdP_T = -beta_TR * dFdf_T
    dFdT_P = dFdT_f + alpha_V * dFdf_T

    # Calculate the unrotated isothermal compressibility
    # and unrotated thermal expansivity tensors
    F = expm(PsiI)
    invF = np.linalg.inv(F)
    LP = np.einsum("ij,kj->ik", dFdP_T, invF)
    beta_T = -0.5 * (LP + LP.T)
    LT = np.einsum("ij,kj->ik", dFdT_P, invF)
    alpha = 0.5 * (LT + LT.T)

    # Calculate the unrotated isothermal compliance
    # tensor in Voigt form.
    S_T = beta_TR * dPsidf_T_Voigt
    for i, j, k in [[0, 1, 2], [1, 2, 0], [0, 2, 1]]:
        S_T[i][j] = 0.5 * (
            -S_T[i][i]
            - S_T[j][j]
            + S_T[k][k]
            + beta_T[i][i]
            + beta_T[j][j]
            - beta_T[k][k]
        )
        S_T[j][i] = S_T[i][j]

        S_T[i][i + 3] = 2.0 * beta_T[j][k] - S_T[j][i + 3] - S_T[k][i + 3]
        S_T[i + 3][i] = S_T[i][i + 3]

    return F, dFdf_T, alpha, S_T


class AnisotropicMineral(Mineral, AnisotropicMaterial):
    """
    A class implementing the anisotropic mineral equation of state described
    in :cite:`Myhill2022`.
    This class is derived from both Mineral and AnisotropicMaterial,
    and inherits most of the methods from these classes.

    Instantiation of an AnisotropicMineral takes three required arguments;
    a reference Mineral (i.e. a standard isotropic mineral which provides
    volume as a function of pressure and temperature), cell_parameters,
    which give the lengths of the molar cell vectors and the angles between
    them (see :func:`~burnman.utils.unitcell.cell_parameters_to_vectors`),
    and an anisotropic parameters object, which should be either a
    4D array of anisotropic parameters or a dictionary of parameters which
    describe the anisotropic behaviour of the mineral.
    For a description of the physical meaning of the parameters in the
    4D array, please refer to the code
    or to the original paper.

    For non-orthotropic materials, the argument frame_convention
    should be set to define the orientation of the reference frame
    relative to the crystallographic axes
    (see :func:`~burnman.utils.unitcell.cell_parameters_to_vectors`).

    If the user chooses to define their parameters as a dictionary,
    they must also provide a function to the psi_function argument
    that describes how to compute the tensors Psi, dPsidf and dPsidPth
    (in Voigt form). The function arguments should be f, Pth and params,
    in that order. The output variables Psi, dPsidf and dPsidth
    must be returned in that order in a tuple. The user should
    also explicitly state whether the material is orthotropic or not
    by supplying a boolean to the orthotropic argument.

    States of the mineral can only be queried after setting the
    pressure and temperature using set_state().

    This class is available as ``burnman.AnisotropicMineral``.

    All the material parameters are expected to be in plain SI units.  This
    means that the elastic moduli should be in Pascals and NOT Gigapascals.
    Additionally, the cell parameters should be in m/(mol formula unit)
    and not in unit cell lengths. To convert unit cell lengths given in
    Angstrom to molar cell parameters you should multiply by 10^(-10) *
    (N_a / Z)^1/3, where N_a is Avogadro's number
    and Z is the number of formula units per unit cell.
    You can look up Z in many places, including www.mindat.org.

    Finally, it is assumed that the unit cell of the anisotropic material
    is aligned in a particular way relative to the coordinate axes
    (the anisotropic_parameters are defined relative to the coordinate axes).
    The crystallographic a-axis is assumed to be parallel to the first
    spatial coordinate axis, and the crystallographic b-axis is assumed to
    be perpendicular to the third spatial coordinate axis.
    """

    def __init__(
        self,
        isotropic_mineral,
        cell_parameters,
        anisotropic_parameters,
        frame_convention=None,
        psi_function=None,
        orthotropic=None,
    ):
        if psi_function is None:
            self.check_standard_parameters(anisotropic_parameters)
            self.anisotropic_params = {"c": anisotropic_parameters}
            self.psi_function = self.standard_psi_function
        else:
            if not isinstance(orthotropic, bool):
                raise Exception(
                    "If the Psi function is provided, "
                    "you must specify whether your material is "
                    "orthotropic as a boolean."
                )
            self.orthotropic = orthotropic
            self.anisotropic_params = anisotropic_parameters
            self.psi_function = psi_function

        if self.orthotropic:
            self.frame_convention = [0, 1, 2]
        else:
            assert (
                len(frame_convention) == 3
            ), "frame_convention must be set for this non-orthotropic mineral"
            self.frame_convention = frame_convention

        Psi_Voigt0 = self.psi_function(0.0, 0.0, self.anisotropic_params)[0]
        if not np.all(Psi_Voigt0 == 0.0):
            raise ValueError(
                "All elements of Psi should evaluate to zero at "
                "standard state. The current array evaluates to: "
                f"{Psi_Voigt0}"
            )

        # cell_vectors is the transpose of the cell tensor M
        self.cell_vectors_0 = cell_parameters_to_vectors(
            cell_parameters, self.frame_convention
        )

        if (
            np.abs(np.linalg.det(self.cell_vectors_0) - isotropic_mineral.params["V_0"])
            > np.finfo(float).eps
        ):
            factor = np.cbrt(
                isotropic_mineral.params["V_0"] / np.linalg.det(self.cell_vectors_0)
            )
            raise Exception(
                "The standard state unit vectors are inconsistent "
                "with the volume. Suggest multiplying each vector length "
                f"by {factor}."
            )

        self.isotropic_mineral = isotropic_mineral
        if "name" in isotropic_mineral.params:
            self.name = isotropic_mineral.params["name"]

        Mineral.__init__(
            self, isotropic_mineral.params, isotropic_mineral.property_modifiers
        )

    def standard_psi_function(self, f, Pth, params):
        # Compute Psi, dPsidPth, dPsidf, needed by most anisotropic properties
        c = params["c"]
        ns = np.arange(c.shape[-1])
        x = c[:, :, 0, :] + c[:, :, 1, :] * f
        dPsidf = c[:, :, 1, :]

        for i in list(range(2, c.shape[2])):
            # non-intuitively, the += operator doesn't simply add in-place,
            # so here we overwrite the arrays with new ones
            x = x + c[:, :, i, :] * np.power(f, float(i)) / float(i)
            dPsidf = dPsidf + c[:, :, i, :] * np.power(f, float(i) - 1.0)

        Psi = np.einsum("ikn, n->ik", x, np.power(Pth, ns))

        dPsidPth = np.einsum(
            "ikn, n->ik", x[:, :, 1:], ns[1:] * np.power(Pth, ns[1:] - 1)
        )

        dPsidf = np.einsum("ikn, n->ik", dPsidf, np.power(Pth, ns))
        return (Psi, dPsidf, dPsidPth)

    @copy_documentation(Material.set_state)
    def set_state(self, pressure, temperature):
        # 1) Compute dPthdf|T
        # relatively large dP needed for accurate estimate of dPthdf

        self.isotropic_mineral.set_state(pressure, temperature)
        V2 = self.isotropic_mineral.V
        KT2 = self.isotropic_mineral.isothermal_bulk_modulus_reuss
        self.isotropic_mineral.set_state_with_volume(V2, self.params["T_0"])
        P1 = self.isotropic_mineral.pressure
        KT1 = self.isotropic_mineral.isothermal_bulk_modulus_reuss
        self.dPthdf = KT1 - KT2
        self.Pth = pressure - P1

        self.isotropic_mineral.set_state(pressure, temperature)
        Mineral.set_state(self, pressure, temperature)

        # 2) Compute other properties needed for anisotropic equation of state
        V = self.V
        V_0 = self.params["V_0"]
        Vrel = V / V_0
        f = np.log(Vrel)

        out = self.psi_function(f, self.Pth, self.anisotropic_params)
        Psi_Voigt, dPsidf_Pth_Voigt, dPsidPth_f_Voigt = out
        Psi_full = voigt_notation_to_compliance_tensor(Psi_Voigt)
        self._PsiI = np.einsum("ijkl, kl", Psi_full, np.eye(3))

        # Change of variables: (f, Pth) -> Psi(f, T)
        aK_T = self.alpha * self.isothermal_bulk_modulus_reuss
        out = convert_f_Pth_to_f_T_derivatives(
            dPsidf_Pth_Voigt, dPsidPth_f_Voigt, aK_T, self.dPthdf
        )
        self._dPsidf_T_Voigt, self._dPsiIdf_T, self._dPsiIdT_f = out

        # Calculate F, thermal expansivity and compliance, dFdT
        if self.orthotropic:
            self._unrotated_F = expm(self._PsiI)
            self._unrotated_alpha = self._dPsiIdT_f + self.alpha * self._dPsiIdf_T
            self._unrotated_S_T_Voigt = (
                self.isothermal_compressibility_reuss * self._dPsidf_T_Voigt
            )
        else:
            out = deformation_gradient_alpha_and_compliance(
                self.alpha,
                self.isothermal_compressibility_reuss,
                self._PsiI,
                self._dPsidf_T_Voigt,
                self._dPsiIdf_T,
                self._dPsiIdT_f,
            )
            (
                self._unrotated_F,
                self._unrotated_dFdf_T,
                self._unrotated_alpha,
                self._unrotated_S_T_Voigt,
            ) = out

    @material_property
    def deformation_gradient_tensor(self):
        """
        :returns: The deformation gradient tensor describing the deformation of the
            mineral from its undeformed state
            (i.e. the state at the reference pressure and temperature).
        :rtype: numpy.array (2D)
        """
        return self._unrotated_F

    @material_property
    def unrotated_cell_vectors(self):
        """
        :returns: The vectors of the cell [m] constructed from one mole
            of formula units after deformation of the mineral from its
            undeformed state (i.e. the state at the reference
            pressure and temperature). See the documentation for the function
            :func:`~burnman.utils.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        :rtype: numpy.array (2D)
        """
        return np.einsum(
            "ij,kj->ki", self.deformation_gradient_tensor, self.cell_vectors_0
        )

    @material_property
    def deformed_coordinate_frame(self):
        """
        :returns: The orientations of the three spatial coordinate axes
            after deformation of the mineral [m]. For orthotropic minerals,
            this is equal to the identity matrix, as hydrostatic stresses only
            induce rotations in monoclinic and triclinic crystals.
        :rtype: numpy.array (2D)
        """
        if self.orthotropic:
            return np.eye(3)
        else:
            c = self.frame_convention
            M_T = self.unrotated_cell_vectors
            Q = np.empty((3, 3))
            Q[c[0]] = M_T[c[0]] / np.linalg.norm(M_T[c[0]])
            Q[c[2]] = np.cross(M_T[c[0]], M_T[c[1]]) / np.linalg.norm(
                np.cross(M_T[c[0]], M_T[c[1]])
            )
            Q[c[1]] = np.cross(Q[c[2]], Q[c[0]])
            return Q

    @material_property
    def rotation_matrix(self):
        """
        :returns: The matrix required to rotate the properties of the deformed
            mineral into the deformed coordinate frame. For orthotropic
            minerals, this is equal to the identity matrix.
        :rtype: numpy.array (2D)
        """
        return self.deformed_coordinate_frame.T

    @material_property
    def cell_vectors(self):
        """
        :returns: The vectors of the cell constructed from one mole
            of formula units [m]. See the documentation for the function
            :func:`~burnman.utils.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        :rtype: numpy.array (2D)
        """
        if self.orthotropic:
            return self.unrotated_cell_vectors
        else:
            vectors = np.einsum(
                "ij, jk->ik", self.unrotated_cell_vectors, self.rotation_matrix
            )
            return vectors

    @material_property
    def cell_parameters(self):
        """
        :returns: The molar cell parameters of the mineral, given in standard form:
            [:math:`a`, :math:`b`, :math:`c`,
            :math:`\\alpha`, :math:`\\beta`, :math:`\\gamma`],
            where the first three floats are the lengths of the vectors in [m]
            defining the cell constructed from one mole of formula units.
            The last three floats are angles between vectors
            (given in radians). See the documentation for the function
            :func:`~burnman.utils.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        :rtype: numpy.array (1D)
        """
        return cell_vectors_to_parameters(self.cell_vectors, self.frame_convention)

    @material_property
    def shear_modulus(self):
        """
        Anisotropic minerals do not (in general) have a single shear modulus.
        This function returns a NotImplementedError. Users should instead
        consider directly querying the elements in the
        isothermal_stiffness_tensor or isentropic_stiffness_tensor.
        """
        raise NotImplementedError(
            "Anisotropic minerals do not have a shear "
            "modulus property. Query "
            "the isentropic or isothermal stiffness "
            "tensors directory, or use"
            "isentropic_shear_modulus_reuss or "
            "isentropic_shear_modulus_voigt."
        )

    @material_property
    def K_T(self):
        """
        Anisotropic minerals do not have a single isothermal bulk modulus.
        This function returns a NotImplementedError. Users should instead
        consider either using isothermal_bulk_modulus_reuss,
        isothermal_bulk_modulus_voigt,
        or directly querying the elements in the isothermal_stiffness_tensor.
        """
        raise NotImplementedError(
            "K_T is not "
            "sufficiently explicit for an "
            "anisotropic mineral. Did you mean "
            "isothermal_bulk_modulus_reuss?"
        )

    isothermal_bulk_modulus_reuss = Mineral.isothermal_bulk_modulus_reuss

    @material_property
    def K_S(self):
        """
        Anisotropic minerals do not have a single isentropic bulk modulus.
        This function returns a NotImplementedError. Users should instead
        consider either using isentropic_bulk_modulus_reuss,
        isentropic_bulk_modulus_voigt (both derived from AnisotropicMineral),
        or directly querying the elements in the isentropic_stiffness_tensor.
        """
        raise NotImplementedError(
            "K_S is not "
            "sufficiently explicit for an "
            "anisotropic mineral. Did you mean "
            "isentropic_bulk_modulus_reuss?"
        )

    isentropic_bulk_modulus_reuss = Mineral.isentropic_bulk_modulus_reuss

    @material_property
    def beta_T(self):
        """
        Anisotropic minerals do not have a single isentropic compressibility.
        This function returns a NotImplementedError. Users should instead
        consider using isothermal_compressibility_tensor,
        isothermal_compressibility_reuss or isothermal_compressibility_voigt
        (all derived from AnisotropicMineral),
        or directly querying the elements in the isothermal_compliance_tensor.
        """
        raise NotImplementedError(
            "beta_T is not "
            "sufficiently explicit for an "
            "anisotropic mineral. Did you mean "
            "isothermal_compressibility_reuss?"
        )

    @material_property
    def beta_S(self):
        """
        Anisotropic minerals do not have a single isentropic compressibility.
        This function returns a NotImplementedError. Users should instead
        consider either using isentropic_compressibility_tensor,
        isentropic_compressibility_reuss or isentropic_compressibility_voigt
        (all derived from AnisotropicMineral),
        or directly querying the elements in the isentropic_compliance_tensor.
        """
        raise NotImplementedError(
            "beta_S is not "
            "sufficiently explicit for an "
            "anisotropic mineral. Did you mean "
            "isentropic_compressibility_reuss?"
        )

    @material_property
    def isothermal_bulk_modulus_voigt(self):
        """
        :returns: The Voigt bound on the isothermal bulk modulus in [Pa].
        :rtype: float
        """
        return np.sum(self.isothermal_stiffness_tensor[:3, :3]) / 9.0

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        :returns: The Reuss bound on the isothermal compressibility in [1/Pa].
        :rtype: float
        """
        return 1.0 / self.isothermal_bulk_modulus_reuss

    beta_T = isothermal_compressibility_reuss

    @material_property
    def isothermal_compressibility_voigt(self):
        """
        :returns: The Voigt bound on the isothermal compressibility in [1/Pa].
        :rtype: float
        """
        return 1.0 / self.isothermal_bulk_modulus_voigt

    @material_property
    def isentropic_bulk_modulus_voigt(self):
        """
        :returns: The Voigt bound on the isentropic bulk modulus in [Pa].
        :rtype: float
        """
        return np.sum(self.isentropic_stiffness_tensor[:3, :3]) / 9.0

    @material_property
    def isentropic_compressibility_reuss(self):
        """
        :returns: The Reuss bound on the isentropic compressibility in [1/Pa].
        :rtype: float
        """
        return 1.0 / self.isentropic_bulk_modulus_reuss

    @material_property
    def isentropic_compressibility_voigt(self):
        """
        :returns: The Voigt bound on the isentropic compressibility in [1/Pa].
        :rtype: float
        """
        return 1.0 / self.isentropic_bulk_modulus_voigt

    @material_property
    def isothermal_compliance_tensor(self):
        """
        :returns: The isothermal compliance tensor [1/Pa]
            in Voigt form (:math:`\\mathbb{S}_{\\text{T} pq}`).
        :rtype: numpy.array (2D)
        """
        if self.orthotropic:
            return self._unrotated_S_T_Voigt
        else:
            R = self.rotation_matrix
            S = voigt_notation_to_compliance_tensor(self._unrotated_S_T_Voigt)
            S_rotated = np.einsum("mi, nj, ok, pl, ijkl->mnop", R, R, R, R, S)
            return contract_compliances(S_rotated)

    @material_property
    def thermal_expansivity_tensor(self):
        """
        :returns: The tensor of thermal expansivities [1/K].
        :rtype: numpy.array (2D)
        """
        if self.orthotropic:
            return self._unrotated_alpha
        else:
            R = self.rotation_matrix
            return np.einsum("mi, nj, ij->mn", R, R, self._unrotated_alpha)

    # Derived properties start here
    @material_property
    def isothermal_stiffness_tensor(self):
        """
        :returns: The isothermal stiffness tensor [Pa]
            in Voigt form (:math:`\\mathbb{C}_{\\text{T} pq}`).
        :rtype: numpy.array (2D)
        """
        return np.linalg.inv(self.isothermal_compliance_tensor)

    @material_property
    def full_isothermal_compliance_tensor(self):
        """
        :returns: The isothermal compliance tensor [1/Pa]
            in standard form (:math:`\\mathbb{S}_{\\text{T} ijkl}`).
        :rtype: numpy.array (4D)
        """
        S_Voigt = self.isothermal_compliance_tensor
        return voigt_notation_to_compliance_tensor(S_Voigt)

    @material_property
    def full_isothermal_stiffness_tensor(self):
        """
        :returns: The isothermal stiffness tensor [Pa]
            in standard form (:math:`\\mathbb{C}_{\\text{T} ijkl}`).
        :rtype: numpy.array (4D)
        """
        CT = self.isothermal_stiffness_tensor
        return voigt_notation_to_stiffness_tensor(CT)

    @material_property
    def full_isentropic_compliance_tensor(self):
        """
        :returns: The isentropic compliance tensor [1/Pa]
            in standard form (:math:`\\mathbb{S}_{\\text{N} ijkl}`).
        :rtype: numpy.array (4D)
        """
        return (
            self.full_isothermal_compliance_tensor
            - np.einsum(
                "ij, kl->ijkl",
                self.thermal_expansivity_tensor,
                self.thermal_expansivity_tensor,
            )
            * self.V
            * self.temperature
            / self.C_p
        )

    @material_property
    def isentropic_compliance_tensor(self):
        """
        :returns: The isentropic compliance tensor [1/Pa]
            in Voigt form (:math:`\\mathbb{S}_{\\text{N} pq}`).
        :rtype: numpy.array (2D)
        """
        S_full = self.full_isentropic_compliance_tensor
        return contract_compliances(S_full)

    @material_property
    def isentropic_stiffness_tensor(self):
        """
        :returns: The isentropic stiffness tensor [Pa]
            in Voigt form (:math:`\\mathbb{C}_{\\text{N} pq}`).
        :rtype: numpy.array (2D)
        """
        return np.linalg.inv(self.isentropic_compliance_tensor)

    @material_property
    def full_isentropic_stiffness_tensor(self):
        """
        :returns: The isentropic stiffness tensor [Pa]
            in standard form (:math:`\\mathbb{C}_{\\text{N} ijkl}`).
        :rtype: numpy.array (4D)
        """
        C_Voigt = self.isentropic_stiffness_tensor
        return voigt_notation_to_stiffness_tensor(C_Voigt)

    @material_property
    def grueneisen_tensor(self):
        """
        :returns: The grueneisen tensor [unitless].
            This is defined by :cite:`BarronMunn1967` as
            :math:`\\mathbb{C}_{\\text{N} ijkl} \\alpha_{kl} V/C_{P}`.
        :rtype: numpy.array (2D)
        """
        return (
            np.einsum(
                "ijkl, kl->ij",
                self.full_isentropic_stiffness_tensor,
                self.thermal_expansivity_tensor,
            )
            * self.molar_volume
            / self.molar_heat_capacity_p
        )

    @material_property
    def molar_heat_capacity_v(self):
        """
        :returns: The isochoric, hydrostatic heat capacity [J/K/mol].
        :rtype: float
        """
        return (
            self.molar_heat_capacity_p
            - self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus_reuss
        )

    @material_property
    def grueneisen_parameter(self):
        """
        :returns: The scalar grueneisen parameter [unitless].
        :rtype: float
        """
        return (
            self.thermal_expansivity
            * self.V
            / (self.isentropic_compressibility_reuss * self.molar_heat_capacity_p)
        )

    @material_property
    def isothermal_compressibility_tensor(self):
        """
        :returns: The isothermal compressibility tensor [1/Pa].
        :rtype: numpy.array (2D)
        """
        return np.einsum(
            "ijkl, kl->ij", self.full_isothermal_compliance_tensor, np.eye(3)
        )

    @material_property
    def isentropic_compressibility_tensor(self):
        """
        :returns: The isentropic compressibility tensor [1/Pa].
        :rtype: numpy.array (2D)
        """
        return np.einsum(
            "ijkl, kl->ij", self.full_isentropic_compliance_tensor, np.eye(3)
        )

    @material_property
    def thermal_stress_tensor(self):
        """
        :returns: The change in stress with temperature at constant strain [Pa/K].
        :rtype: numpy.array (2D)
        """
        pi = -np.einsum(
            "ijkl, kl",
            self.full_isothermal_stiffness_tensor,
            self.thermal_expansivity_tensor,
        )
        return pi

    @material_property
    def molar_isometric_heat_capacity(self):
        """
        :returns: The molar heat capacity at constant strain [J/K/mol].
        :rtype: float
        """
        alpha = self.thermal_expansivity_tensor
        pi = self.thermal_stress_tensor
        C_isometric = (
            self.molar_heat_capacity_p
            + self.V * self.temperature * np.einsum("ij, ij", alpha, pi)
        )
        return C_isometric

    def check_standard_parameters(self, anisotropic_parameters):
        if not np.all(anisotropic_parameters[:, :, 0, 0] == 0):
            raise Exception(
                "anisotropic_parameters_pqmn should be set to " "zero for all m = n = 0"
            )

        sum_ijij_block = np.sum(anisotropic_parameters[:3, :3, :, :], axis=(0, 1))

        if np.abs(sum_ijij_block[1, 0] - 1.0) > 1.0e-5:
            raise Exception(
                "The sum of the upper 3x3 pq-block of "
                "anisotropic_parameters_pqmn must equal "
                "1 for m=1, n=0 for consistency with the volume. "
                f"Value is {sum_ijij_block[1, 0]}"
            )

        for m in range(2, len(sum_ijij_block)):
            if np.abs(sum_ijij_block[m, 0]) > 1.0e-10:
                raise Exception(
                    "The sum of the upper 3x3 pq-block of "
                    "anisotropic_parameters_pqmn must equal 0 for "
                    f"m={m}, n=0 for consistency with the volume. "
                    f"Value is {sum_ijij_block[m, 0]}"
                )

        for m in range(len(sum_ijij_block)):
            for n in range(1, len(sum_ijij_block[0])):
                if np.abs(sum_ijij_block[m, n]) > 1.0e-10:
                    raise Exception(
                        "The sum of the upper 3x3 pq-block of "
                        "anisotropic_parameters_pqmn must equal "
                        f"0 for m={m}, n={n} for "
                        "consistency with the volume. "
                        f"Value is {sum_ijij_block[m, n]}"
                    )

        for m in range(len(sum_ijij_block)):
            for n in range(len(sum_ijij_block[0])):
                a = anisotropic_parameters[:, :, m, n]
                if not np.allclose(a, a.T, rtol=1.0e-8, atol=1.0e-8):
                    raise Exception(
                        f"The anisotropic_parameters_pq{m}{n} must be symmetric."
                    )

        if cond(anisotropic_parameters[:, :, 1, 0]) > 1 / np.finfo(float).eps:
            raise Exception("anisotropic_parameters[:, :, 1, 0] is singular")

        sum_lower_left_block = np.sum(anisotropic_parameters[3:, :3, :, :], axis=1)

        self.orthotropic = True
        for i, s in enumerate(sum_lower_left_block):
            if not np.all(np.abs(s) < 1.0e-10):
                self.orthotropic = False



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


class EquationOfState(object):
    """
    This class defines the interface for an equation of state
    that a mineral uses to determine its properties at a
    given :math:`P, T`.  In order define a new equation of state, you
    should define these functions.

    All functions should accept and return values in SI units.

    In general these functions are functions of pressure,
    temperature, and volume, as well as a "params" object,
    which is a Python dictionary that stores the material
    parameters of the mineral, such as reference volume,
    Debye temperature, reference moduli, etc.

    The functions for volume and density are just functions
    of temperature, pressure, and "params"; after all, it
    does not make sense for them to be functions of volume or density.
    """

    def volume(self, pressure, temperature, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Molar volume of the mineral :math:`[m^3]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def pressure(self, temperature, volume, params):
        """
        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Pressure of the mineral, including cold and thermal parts
            :math:`[m^3]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def density(self, volume, params):
        """
        Calculate the density of the mineral :math:`[kg/m^3]`.
        The params object must include a "molar_mass" field.

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Density of the mineral. :math:`[kg/m^3]`
        :rtype: float
        """
        return params["molar_mass"] / volume

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Grueneisen parameter of the mineral. :math:`[unitless]`
        :rtype: float
        """
        raise NotImplementedError("")

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Isothermal bulk modulus of the mineral. :math:`[Pa]`
        :rtype: float
        """
        raise NotImplementedError("")

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Adiabatic bulk modulus of the mineral. :math:`[Pa]`
        :rtype: float
        """
        raise NotImplementedError("")

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Shear modulus of the mineral. :math:`[Pa]`
        :rtype: float
        """
        raise NotImplementedError("")

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Heat capacity at constant volume of the mineral. :math:`[J/K/mol]`
        :rtype: float
        """
        raise NotImplementedError("")

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Heat capacity at constant pressure of the mineral. :math:`[J/K/mol]`
        :rtype: float
        """
        raise NotImplementedError("")

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Thermal expansivity of the mineral. :math:`[1/K]`
        :rtype: float
        """
        raise NotImplementedError("")

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Gibbs energy of the mineral :math:`[J/mol]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Helmholtz energy of the mineral :math:`[J/mol]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def entropy(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Entropy of the mineral :math:`[J/K/mol]`.
        :rtype: float
        """

        raise NotImplementedError("")

    def enthalpy(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Enthalpy of the mineral :math:`[J/mol]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        :param pressure: Pressure at which to evaluate the equation of state
            :math:`[Pa]`.
        :type pressure: float

        :param temperature: Temperature at which to evaluate the equation of state
            :math:`[K]`.
        :type temperature: float

        :param volume: Molar volume of the mineral. For consistency this should be
            calculated using :func:`volume` :math:`[m^3]`.
        :type volume: float

        :param params: Dictionary containing material parameters required by
            the equation of state.
        :type params: dict

        :returns: Internal energy of the mineral :math:`[J/mol]`.
        :rtype: float
        """
        raise NotImplementedError("")

    def validate_parameters(self, params):
        """
        The params object is just a dictionary associating mineral physics parameters
        for the equation of state.
        Different equation of states can have different parameters,
        and the parameters may have ranges of validity.  The intent of this function is
        twofold. First, it can check for the existence of the parameters that the
        equation of state needs, and second, it can check whether the parameters
        have reasonable values.  Unreasonable values will frequently be due
        to unit issues (e.g., supplying bulk moduli in GPa instead of Pa).
        In the base class this function does nothing, and an equation of state
        is not required to implement it.  This function will
        not return anything, though it may raise warnings or errors.

        :param params: Dictionary containing material parameters required by the
            equation of state.
        :type params: dict
        """
        pass



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np

from .mineral import Mineral, material_property
from .solution import Solution
from .solutionmodel import MechanicalSolution


class CombinedMineral(Mineral):
    """
    This is the base class for endmembers constructed from a
    linear combination of other minerals.

    Instances of this class should be initialised with a
    list of Mineral instances, a second list containing the
    number of moles of each mineral, and (optionally) a
    third list containing three floats describing a
    free energy adjustment which is linear in pressure and temperature
    (i.e. a constant energy [J/mol], entropy [J/K/mol]
    and volume adjustment [J/Pa/mol or m^3/mol]).

    For example, a crude approximation to a bridgmanite model might be
    bdg = CombinedMineral([per, stv], [1.0, 1.0], [-15.e3, 0., 0.])

    This class is available as :class:`burnman.CombinedMineral`.
    """

    def __init__(
        self,
        mineral_list,
        molar_amounts,
        free_energy_adjustment=[],
        name="User-created endmember",
    ):
        model = MechanicalSolution(endmembers=[[m, ""] for m in mineral_list])
        self.mixture = Solution(solution_model=model, molar_fractions=molar_amounts)

        # Remove elements from the chemical formula if they have
        # negligible concentrations
        for key, value in list(self.mixture.formula.items()):
            if np.abs(value) < 1.0e-10:
                self.mixture.formula.pop(key)

        self.params = {
            "name": name,
            "formula": self.mixture.formula,
            "equation_of_state": "combined",
            "molar_mass": self.mixture.molar_mass,
            "n": sum(self.mixture.formula.values()),
        }

        if free_energy_adjustment != []:
            assert len(free_energy_adjustment) == 3
            dE, dS, dV = free_energy_adjustment
            self.property_modifiers = [
                ["linear", {"delta_E": dE, "delta_S": dS, "delta_V": dV}]
            ]

        Mineral.__init__(self)

    def set_state(self, pressure, temperature):
        self.mixture.set_state(pressure, temperature)
        Mineral.set_state(self, pressure, temperature)

    @material_property
    def molar_gibbs(self):
        """
        Returns Gibbs free energy of the mineral [J]
        Aliased with self.gibbs
        """
        return self.mixture.molar_gibbs + self._property_modifiers["G"]

    @material_property
    def _molar_volume_unmodified(self):
        return self.mixture.molar_volume

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the mineral [m^3/mol]
        Aliased with self.V
        """
        return self.mixture.molar_volume + self._property_modifiers["dGdP"]

    @material_property
    def molar_entropy(self):
        """
        Returns entropy of the mineral [J]
        Aliased with self.S
        """
        return self.mixture.molar_entropy - self._property_modifiers["dGdT"]

    @material_property
    def isothermal_bulk_modulus_reuss(self):
        """
        Returns isothermal bulk modulus of the mineral [Pa]
        Aliased with self.K_T
        """
        K_T_orig = self.mixture.isothermal_bulk_modulus_reuss

        return self.molar_volume / (
            (self._molar_volume_unmodified / K_T_orig)
            - self._property_modifiers["d2GdP2"]
        )

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the mineral [Pa]
        Aliased with self.G
        """
        return self.mixture.shear_modulus

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha) of the mineral [1/K]
        Aliased with self.alpha
        """
        return (
            (self.mixture.thermal_expansivity * self._molar_volume_unmodified)
            + self._property_modifiers["d2GdPdT"]
        ) / self.molar_volume

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns heat capacity at constant pressure of the mineral [J/K/mol]
        Aliased with self.C_p
        """
        return (
            self.mixture.molar_heat_capacity_p
            - self.temperature * self._property_modifiers["d2GdT2"]
        )

    """
    Properties from mineral parameters,
    Legendre transformations
    or Maxwell relations
    """

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the mineral [kg/mol]
        """
        return self.mixture.molar_mass

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the mineral
        """
        return self.mixture.formula

    @material_property
    def density(self):
        """
        Returns density of the mineral [kg/m^3]
        Aliased with self.rho
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol]
        Aliased with self.energy
        """
        return (
            self.molar_gibbs
            - self.pressure * self.molar_volume
            + self.temperature * self.molar_entropy
        )

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the mineral [J/mol]
        Aliased with self.helmholtz
        """
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the mineral [J/mol]
        Aliased with self.H
        """
        return self.molar_gibbs + self.temperature * self.molar_entropy

    @material_property
    def isentropic_bulk_modulus_reuss(self):
        """
        Returns adiabatic bulk modulus of the mineral [Pa]
        Aliased with self.K_S
        """
        if self.temperature < 1.0e-10:
            return self.isothermal_bulk_modulus_reuss
        else:
            return (
                self.isothermal_bulk_modulus_reuss
                * self.molar_heat_capacity_p
                / self.molar_heat_capacity_v
            )

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        Returns isothermal compressibility of the mineral
        (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.K_T
        """
        return 1.0 / self.isothermal_bulk_modulus_reuss

    @material_property
    def isentropic_compressibility_reuss(self):
        """
        Returns adiabatic compressibility of the mineral
        (or inverse adiabatic bulk modulus) [1/Pa]
        Aliased with self.K_S
        """
        return 1.0 / self.isentropic_bulk_modulus_reuss

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the mineral [m/s]
        Aliased with self.v_p
        """
        return np.sqrt(
            (self.isentropic_bulk_modulus_reuss + 4.0 / 3.0 * self.shear_modulus)
            / self.density
        )

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the mineral [m/s]
        Aliased with self.v_phi
        """
        return np.sqrt(self.isentropic_bulk_modulus_reuss / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the mineral [m/s]
        Aliased with self.v_s
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the mineral [unitless]
        Aliased with self.gr
        """
        if self.temperature < 1.0e-12:
            return 0.0
        else:
            return (
                self.thermal_expansivity
                * self.isothermal_bulk_modulus_reuss
                * self.molar_volume
                / self.molar_heat_capacity_v
            )

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the mineral [J/K/mol]
        Aliased with self.C_v
        """
        return (
            self.molar_heat_capacity_p
            - self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus_reuss
        )



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from copy import deepcopy
from collections import namedtuple
from .linear_fitting import weighted_constrained_least_squares
from ..utils.chemistry import dictionarize_formula
from ..utils.chemistry import process_solution_chemistry
from ..classes.solution import Solution


class DummyCompositionSolution(Solution):
    """
    This is a dummy base class for a solution object to
    facilitate composition fitting when no solution model has
    been prepared. The model is initialized with appropriate
    chemical formulae for each endmember, and can do all basic
    compositional processing that doesn't involve any material
    properties.

    :param endmember_element_formulae: Formulae for each of the independent endmembers.
        e.g. ['Mg2SiO4', 'Fe2SiO4'].
    :type endmember_element_formulae: list of str
    :param endmember_site_formulae: Site formulae for each of the independent endmembers,
        in the same order as endmember_element_formulae. e.g. ['[Mg]2SiO4', '[Fe]2SiO4'].
    :type endmember_site_formulae: list of str
    """

    def __init__(self, endmember_element_formulae, endmember_site_formulae):
        self.endmember_formulae = [
            dictionarize_formula(f) for f in endmember_element_formulae
        ]
        self.solution_model = type(
            "Dimension", (object,), {"formulas": endmember_site_formulae}
        )()
        self.solution_model.endmembers = [None for f in endmember_site_formulae]
        process_solution_chemistry(self.solution_model)


def fit_composition_to_solution(
    solution,
    fitted_variables,
    variable_values,
    variable_covariances,
    variable_conversions=None,
    normalize=True,
):
    """
    Takes a Solution object and a set of variable names and
    associates values and covariances and finds the molar fractions of the
    solution which provide the best fit (in a least-squares sense)
    to the variable values.

    The fitting applies appropriate non-negativity constraints
    (i.e. no species can have a negative occupancy on a site).

    :param solution: The solution to use in the fitting procedure.
    :type solution: :class:`burnman.Solution`

    :param fitted_variables: A list of the variables used to find the best-fit
        molar fractions of the solution. These should either be elements
        such as "Fe", site_species such as "Fef_B" which would correspond to a
        species labelled Fef on the second site,
        or user-defined variables which are arithmetic sums of
        elements and/or site_species defined in "variable_conversions".
    :type fitted_variables: list of str

    :param variable_values: Numerical values of the fitted variables.
        These should be given as amounts; they do not need to be normalized.
    :type variable_values: numpy.array

    :param variable_covariances: Covariance matrix of the variables.
    :type variable_covariances: 2D numpy.array

    :param variable_conversions: A dictionary converting any user-defined variables
        into an arithmetic sum of element and site-species amounts. For example,
        {'Mg_equal': {'Mg_A': 1., 'Mg_B': -1.}}, coupled with Mg_equal = 0
        would impose a constraint that the amount of Mg would be equal on
        the first and second site in the solution.
    :type variable_conversions: dict of dict, or None

    :param normalize: If True, normalizes the optimized molar fractions to sum to unity.
    :type normalize: bool

    :returns: Optimized molar fractions, corresponding covariance matrix and the
        weighted residual.
    :rtype: tuple of 1D numpy.array, 2D numpy.array and float
    """

    solution_variables = deepcopy(solution.elements)
    solution_variables.extend(solution.solution_model.site_names)

    solution_matrix = np.hstack(
        (solution.stoichiometric_matrix, solution.solution_model.endmember_noccupancies)
    )

    n_vars = len(fitted_variables)
    n_mbrs, n_sol_vars = solution_matrix.shape

    if variable_conversions is not None:
        solution_matrix = np.hstack(
            (
                solution_matrix,
                np.zeros((solution_matrix.shape[0], len(variable_conversions))),
            )
        )

        for i, (new_var, conversion_dict) in enumerate(variable_conversions.items()):
            assert new_var not in solution_variables
            solution_variables.append(new_var)

            for var in conversion_dict.keys():
                solution_matrix[:, n_sol_vars + i] += solution_matrix[
                    :, solution_variables.index(var)
                ]

    # Now, construct A using the fitted variables
    A = np.zeros((n_vars, solution_matrix.shape[0]))
    for i, var in enumerate(fitted_variables):
        A[i, :] = solution_matrix[:, solution_variables.index(var)]

    b = variable_values
    Cov_b = variable_covariances

    # Define the constraints
    # Ensure that element abundances / site occupancies
    # are exactly equal to zero if the user specifies that
    # they are equal to zero.
    S, S_index = np.unique(A, axis=0, return_index=True)
    S = np.array(
        [
            s
            for i, s in enumerate(S)
            if np.abs(b[S_index[i]]) < 1.0e-10 and any(np.abs(s) > 1.0e-10)
        ]
    )
    equality_constraints = [S, np.zeros(len(S))]

    # Ensure all site occupancies are non-negative
    T = np.array(
        [
            -t
            for t in np.unique(solution.solution_model.endmember_occupancies.T, axis=0)
            if any(np.abs(t) > 1.0e-10)
        ]
    )
    inequality_constraints = [T, np.zeros(len(T))]

    popt, pcov, res = weighted_constrained_least_squares(
        A, b, Cov_b, equality_constraints, inequality_constraints
    )

    if normalize:
        sump = sum(popt)
        popt /= sump
        pcov /= sump * sump
        res /= sump

    # Convert the variance-covariance matrix from endmember amounts to
    # endmember proportions
    dpdx = (np.eye(n_mbrs) - popt).T  # = (1. - p[i] if i == j else -p[i])
    pcov = dpdx.dot(pcov).dot(dpdx.T)
    return (popt, pcov, res)


def fit_phase_proportions_to_bulk_composition(phase_compositions, bulk_composition):
    """
    Performs weighted constrained least squares on a set of phase compositions
    to find the amount of those phases that best-fits a given bulk composition.

    The fitting applies appropriate non-negativity constraints
    (i.e. no phase can have a negative abundance in the bulk).

    :param phase_compositions: The composition of each phase.
        Can be in weight or mole amounts.
    :type phase_compositions: 2D numpy.array

    :param bulk_composition: The bulk composition of the composite.
        Must be in the same units as the phase compositions.
    :type bulk_composition: numpy.array

    :returns: Optimized molar fractions, corresponding covariance matrix and the
        weighted residual.
    :rtype: tuple of 1D numpy.array, 2D numpy.array and float
    """

    n_phases = len(phase_compositions[0])
    inequality_constraints = [-np.eye(n_phases), np.zeros(n_phases)]
    popt, pcov, res = weighted_constrained_least_squares(
        phase_compositions, bulk_composition, None, None, inequality_constraints
    )
    return (popt, pcov, res)



import numpy as np


def pressure_to_pressure(
    old_calibrant, new_calibrant, pressure, temperature, PT_covariance=None
):
    """
    Convert from pressure defined by one calibrated equation of
    state of a material to pressure defined by an independent
    calibration of the same material.

    :param old_calibrant: The original calibration used to estimate the pressure
    :type old_calibrant: :class:`burnman.Calibrant`

    :param new_calibrant: The new calibration from which the pressure is desired
    :type new_calibrant: :class:`burnman.Calibrant`

    :param pressure: The pressure calculated using the old calibration
    :type pressure: float

    :param temperature: The temperature of the material
    :type temperature: float

    :param PT_covariance: The pressure-temperature variance-covariance matrix
    :type PT_covariance: 2x2 numpy.array [optional]

    :returns: The pressure of the calibrant [Pa] (float) and a 2x2 numpy array
        (if the PT_covariance is provided) containing the
        pressure-temperature variance-covariance matrix.

    :rtype: tuple
    """

    if PT_covariance is None:
        V = old_calibrant.volume(pressure, temperature)
        P = new_calibrant.pressure(V, temperature)
        return P
    else:
        V, var_VPT = old_calibrant.volume(pressure, temperature, PT_covariance)
        VT_covariance = var_VPT[np.ix_([0, 2], [0, 2])]
        P, var_PVT = new_calibrant.pressure(V, temperature, VT_covariance)
        PT_covariance = var_PVT[np.ix_([0, 2], [0, 2])]
        return P, PT_covariance



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from .. import constants


def molar_volume_from_unit_cell_volume(unit_cell_v, z):
    """
    Converts a unit cell volume from Angstroms^3 per unitcell,
    to m^3/mol.

    :param unit_cell_v: Unit cell volumes [A^3/unit cell].
    :type unit_cell_v: float

    :param z: Number of formula units per unit cell.
    :type z: float

    :returns: Volume [m^3/mol]
    :rtype: float
    """
    V = unit_cell_v * constants.Avogadro / 1.0e30 / z
    return V


def cell_parameters_to_vectors(cell_parameters, frame_convention):
    """
    Converts cell parameters to unit cell vectors.

    :param cell_parameters: An array containing the three lengths of the
        unit cell vectors [m], and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.
    :type cell_parameters: numpy.array (1D)

    :param frame_convention: A list (c) defining the reference frame
        convention.  This function dictates that the c[0]th cell vector
        is colinear with the c[0]th axis, the c[1]th cell vector is
        perpendicular to the c[2]th axis,
        and the c[2]th cell vector is defined to give a right-handed
        coordinate system.
        In common crystallographic shorthand, x[c[0]] // a[c[0]],
        x[c[2]] // a[c[2]]^* (i.e. perpendicular to a[c[0]] and a[c[1]]).

    :type frame_convention: list of three integers

    :returns: The three vectors defining the parallelopiped cell [m].

    :rtype: numpy.array (2D)
    """
    lengths = cell_parameters[:3]
    angles_deg = cell_parameters[3:]
    angles = np.radians(angles_deg)

    c = frame_convention

    n2 = (np.cos(angles[c[0]]) - np.cos(angles[c[2]]) * np.cos(angles[c[1]])) / np.sin(
        angles[c[2]]
    )
    MT = np.zeros((3, 3))
    MT[c[0], c[0]] = lengths[c[0]]
    MT[c[1], c[0]] = lengths[c[1]] * np.cos(angles[c[2]])
    MT[c[1], c[1]] = lengths[c[1]] * np.sin(angles[c[2]])
    MT[c[2], c[0]] = lengths[c[2]] * np.cos(angles[c[1]])
    MT[c[2], c[1]] = lengths[c[2]] * n2
    MT[c[2], c[2]] = lengths[c[2]] * np.sqrt(np.sin(angles[c[1]]) ** 2 - n2**2)

    return MT


def cell_vectors_to_parameters(vectors, frame_convention):
    """
    Converts unit cell vectors to cell parameters.

    :param vectors: The three vectors defining the parallelopiped cell [m].
    :type vectors: numpy.array (2D)

    :param frame_convention: A list (c) defining the reference frame
        convention. This function dictates that the c[0]th cell vector
        is colinear with the c[0]th axis, the c[1]th cell vector is
        perpendicular to the c[2]th axis,
        and the c[2]th cell vector is defined to give a right-handed
        coordinate system.
        In common crystallographic shorthand, x[c[0]] // a[c[0]],
        x[c[2]] // a[c[2]]^* (i.e. perpendicular to a[c[0]] and a[c[1]]).

    :type frame_convention: list of three integers

    :returns: An array containing the three lengths of the unit cell vectors [m],
        and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.
    :rtype: numpy.array (1D)
    """

    c = frame_convention
    assert np.abs(vectors[c[0], c[1]]) < np.finfo(float).eps
    assert np.abs(vectors[c[0], c[2]]) < np.finfo(float).eps
    assert np.abs(vectors[c[1], c[2]]) < np.finfo(float).eps

    lengths = np.empty(3)
    angles = np.empty(3)

    lengths[c[0]] = vectors[c[0], c[0]]
    lengths[c[1]] = np.sqrt(
        np.power(vectors[c[1], c[0]], 2.0) + np.power(vectors[c[1], c[1]], 2.0)
    )
    lengths[c[2]] = np.sqrt(
        np.power(vectors[c[2], c[0]], 2.0)
        + np.power(vectors[c[2], c[1]], 2.0)
        + np.power(vectors[c[2], c[2]], 2.0)
    )

    angles[c[2]] = np.arccos(vectors[c[1], c[0]] / lengths[c[1]])
    angles[c[1]] = np.arccos(vectors[c[2], c[0]] / lengths[c[2]])
    angles[c[0]] = np.arccos(
        vectors[c[2], c[1]] / lengths[c[2]] * np.sin(angles[c[2]])
        + np.cos(angles[c[2]]) * np.cos(angles[c[1]])
    )

    angles_deg = np.degrees(angles)
    return np.array([*lengths, *angles_deg])



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# This module provides higher level chemistry-related functions.

from __future__ import absolute_import
import numpy as np
from scipy.optimize import fsolve
import itertools
from sympy import Rational

from .. import constants

# Import common lower level functions for backwards compatibility
from ..classes.polytope import MaterialPolytope
from ..utils.chemistry import dictionarize_formula, formula_mass
from ..utils.chemistry import formula_to_string, site_occupancies_to_strings
from ..utils.chemistry import compositional_array
from ..utils.chemistry import reaction_matrix_as_strings


def fugacity(standard_material, assemblage):
    """
    Calculates the fugacity of a standard material in another assemblage.

    .. note:: set_method and set_state should already have been
        used on both assemblages.

    :param standard_material: Standard material for which to calculate the fugacity.
        The material must have a formula as a dictionary parameter.
    :type assemblage: :class:`burnman.Material`

    :param assemblage: Assemblage for which to calculate the fugacity.
    :type assemblage: :class:`burnman.Composite`

    :returns: Value of the fugacity of the component with respect to
        the standard material.
    :rtype: float
    """
    component_formula = standard_material.params["formula"]
    chemical_potential = assemblage.chemical_potential([component_formula])[0]

    fugacity = np.exp(
        (chemical_potential - standard_material.gibbs)
        / (constants.gas_constant * assemblage.temperature)
    )
    return fugacity


def relative_fugacity(component_formula, assemblage, reference_assemblage):
    """
    Calculates the fugacity of a chemical component in one assemblage
    relative to another one.

    .. note:: set_method and set_state should already have been
        used on both assemblages.

    :param component_formula: Chemical formula for which to compute the
        relative fugacity.
    :type component_formula: dictionary

    :param assemblage: Assemblage for which to calculate the fugacity.
    :type assemblage: :class:`burnman.Composite`

    :param reference_assemblage: Reference assemblage against which to
        measure the fugacity.
    :type reference_assemblage: :class:`burnman.Composite`

    :returns: Value of the fugacity of the component in the assemblage
        with respect to the reference_assemblage.
    :rtype: float
    """
    chemical_potential = assemblage.chemical_potential([component_formula])[0]
    reference_chemical_potential = reference_assemblage.chemical_potential(
        [component_formula]
    )[0]

    relative_fugacity = np.exp(
        (chemical_potential - reference_chemical_potential)
        / (constants.gas_constant * assemblage.temperature)
    )
    return relative_fugacity


def equilibrium_pressure(
    minerals, stoichiometry, temperature, pressure_initial_guess=1.0e5
):
    """
    Given a list of minerals, their reaction stoichiometries
    and a temperature of interest, compute the
    equilibrium pressure of the reaction.

    :param minerals: List of minerals involved in the reaction.
    :type minerals: list of :class:`burnman.Mineral`

    :param stoichiometry: Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol].
    :type stoichiometry: list of floats

    :param temperature: Temperature of interest [K].
    :type temperature: float

    :param pressure_initial_guess: Initial pressure guess [Pa].
    :type pressure_initial_guess: float

    :returns: The equilibrium pressure of the reaction [Pa].
    :rtype: float
    """

    def eqm(P, T):
        gibbs = 0.0
        for i, mineral in enumerate(minerals):
            mineral.set_state(P[0], T)
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    pressure = fsolve(eqm, [pressure_initial_guess], args=(temperature))[0]

    return pressure


def equilibrium_temperature(
    minerals, stoichiometry, pressure, temperature_initial_guess=1000.0
):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    :param minerals: List of minerals involved in the reaction.
    :type minerals: list of :class:`burnman.Mineral`

    :param stoichiometry: Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol].
    :type stoichiometry: list of floats

    :param pressure: Pressure of interest [Pa].
    :type pressure: float

    :param temperature_initial_guess: Initial temperature guess [K].
    :type temperature_initial_guess: float

    :returns: The equilibrium temperature of the reaction [K].
    :rtype: float
    """

    def eqm(T, P):
        gibbs = 0.0
        for i, mineral in enumerate(minerals):
            mineral.set_state(P, T[0])
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    temperature = fsolve(eqm, [temperature_initial_guess], args=(pressure))[0]

    return temperature


def invariant_point(
    minerals_r1,
    stoichiometry_r1,
    minerals_r2,
    stoichiometry_r2,
    pressure_temperature_initial_guess=[1.0e9, 1000.0],
):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    :param minerals: List of minerals involved in the reaction.
    :type minerals: list of :class:`burnman.Mineral`

    :param stoichiometry: Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol].
    :type stoichiometry: list of floats

    :param pressure: Pressure of interest [Pa].
    :type pressure: float

    :param temperature_initial_guess: Initial temperature guess [K].
    :type temperature_initial_guess: float

    :returns: The equilibrium temperature of the reaction [K].
    :rtype: float
    """

    def eqm(PT):
        P, T = PT
        gibbs_r1 = 0.0
        for i, mineral in enumerate(minerals_r1):
            mineral.set_state(P, T)
            gibbs_r1 = gibbs_r1 + mineral.gibbs * stoichiometry_r1[i]
        gibbs_r2 = 0.0
        for i, mineral in enumerate(minerals_r2):
            mineral.set_state(P, T)
            gibbs_r2 = gibbs_r2 + mineral.gibbs * stoichiometry_r2[i]
        return [gibbs_r1, gibbs_r2]

    pressure, temperature = fsolve(eqm, pressure_temperature_initial_guess)
    return pressure, temperature


def hugoniot(mineral, P_ref, T_ref, pressures, reference_mineral=None):
    """
    Calculates the temperatures (and volumes) along a Hugoniot
    as a function of pressure according to the Hugoniot equation
    U2-U1 = 0.5*(p2 - p1)(V1 - V2) where U and V are the
    internal energies and volumes (mass or molar) and U = F + TS

    :param mineral:  Mineral for which the Hugoniot is to be calculated.
    :type mineral: :class:`burnman.Mineral`

    :param P_ref: Reference pressure [Pa]
    :type P_ref: float

    :param T_ref:  Reference temperature [K]
    :type T_ref: float

    :param pressures: Set of pressures [Pa] for which the Hugoniot temperature
        and volume should be calculated.
    :type pressures: numpy.array of floats

    :param reference_mineral: Mineral which is stable at the reference conditions
        Provides an alternative U_0 and V_0 when the reference
        mineral transforms to the mineral of interest at some
        (unspecified) pressure.
    :type reference_mineral: :class:`burnman.Mineral`

    :returns: The Hugoniot temperatures and volumes at the given pressures.
    :rtype: tuple of numpy.arrays
    """

    def Ediff(T, mineral, P, P_ref, U_ref, V_ref):
        mineral.set_state(P, T[0])
        U = mineral.helmholtz + T[0] * mineral.S
        V = mineral.V

        return (U - U_ref) - 0.5 * (P - P_ref) * (V_ref - V)

    if reference_mineral is None:
        reference_mineral = mineral

    reference_mineral.set_state(P_ref, T_ref)
    U_ref = reference_mineral.helmholtz + T_ref * reference_mineral.S
    V_ref = reference_mineral.V

    temperatures = np.empty_like(pressures)
    volumes = np.empty_like(pressures)

    for i, P in enumerate(pressures):
        temperatures[i] = fsolve(
            Ediff, [T_ref], args=(mineral, P, P_ref, U_ref, V_ref)
        )[0]
        volumes[i] = mineral.V

    return temperatures, volumes


def reactions_from_stoichiometric_matrix(stoichiometric_matrix):
    """
    Returns a list of all the balanced reactions between compounds
    of fixed chemical composition. Includes both the forward and
    reverse reactions
    (so there will always be an even number of reactions).

    :param stoichiometric_matrix: An array of the stoichiometric
        (molar) amounts of component j in compound i.
    :type stoichiometric_matrix: 2D numpy array

    :returns: An array of the stoichiometric (molar) amounts of
        compound j in reaction i.
    :rtype: 2D numpy array
    """
    n_components = stoichiometric_matrix.shape[1]

    equalities = np.concatenate(
        ([np.zeros(n_components)], np.array(stoichiometric_matrix).astype(float))
    ).T

    polys = [
        MaterialPolytope(equalities, np.diag(v).astype(float))
        for v in itertools.product(*[[-1, 1]] * len(equalities[0]))
    ]
    reactions = []
    for p in polys:
        v = np.array(
            [
                [Rational(value).limit_denominator(1000000) for value in v]
                for v in p.raw_vertices
            ],
            dtype=float,
        )

        if v is not []:
            reactions.extend(v)

    reactions = np.unique(np.array(reactions), axis=0)
    reactions = np.array(
        [[Rational(value).limit_denominator(1000000) for value in v] for v in reactions]
    )

    assert np.max(reactions[:-1, 0]) == 0
    assert np.max(reactions[-1, 1:]) == 0
    reactions = reactions[:-1, 1:]
    return reactions


def reactions_from_formulae(formulae, compound_names, return_strings=True):
    """
    Returns a list of all the balanced reactions between compounds
    of fixed chemical composition. Includes both the forward and
    reverse reactions
    (so there will always be an even number of reactions).

    :param formulae: List of the chemical formulae, either as strings or
        as a list of dictionaries of elements.
    :type formulae: list of dictionaries or list of strings

    :param compound_names: List of the compound names in the formula list.
    :type compound_names: list of strings

    :param return_strings: Whether to return the reactions as strings or array.
    :type return_strings: bool

    :returns: Either a 2D array of the stoichiometric (molar) amounts of
        compound j in reaction i, or a list of strings.
        The parameter compound_names is only used if strings
        are requested.
    :rtype: 2D numpy array or list of strings
    """
    if isinstance(formulae[0], str):
        dict_formulae = [dictionarize_formula(f) for f in formulae]
    else:
        dict_formulae = formulae
    stoichiometric_matrix, elements = compositional_array(dict_formulae)
    R = reactions_from_stoichiometric_matrix(stoichiometric_matrix)
    if return_strings:
        return reaction_matrix_as_strings(R, compound_names)
    else:
        return R



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences.
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from scipy.special import gamma, gammainc, exp1
from . import equation_of_state as eos
import warnings
import numpy as np


# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:

    def jit(fn):
        return fn


def generalised_gammainc(a, x1, x2):
    """
    An implementation of the generalised incomplete gamma
    function. Computed using the relationship with the regularised
    lower incomplete gamma function (scipy.special.gammainc).
    Uses the recurrence relation wherever a<0.

    We could have used mpmath.gammainc(a, x1, x2) directly,
    but it is significantly slower than this implementation.
    """
    n = int(-np.floor(a))
    if n > 0:
        a = a + n
        u_gamma = (
            exp1(x1) - exp1(x2)
            if np.abs(a) < 1.0e-12
            else (gammainc(a, x2) - gammainc(a, x1)) * gamma(a)
        )

        esubx1 = np.exp(-x1)
        esubx2 = np.exp(-x2)
        for _ in range(n):
            a = a - 1.0
            u_gamma = (
                u_gamma + np.power(x2, a) * esubx2 - np.power(x1, a) * esubx1
            ) / a
        return u_gamma
    else:
        return (gammainc(a, x2) - gammainc(a, x1)) * gamma(a)


@jit(nopython=True)
def make_params(K_0, Kp_0, Kp_inf, Kdp_0):
    dKpdlnV_zero = -Kdp_0 * K_0
    c = Kp_inf
    a = dKpdlnV_zero / (Kp_0 - Kp_inf)
    b = (Kp_0 - Kp_inf) / a
    return a, b, c


class SPOCK(eos.EquationOfState):
    """
    Class for the Scaled Power Of Compression K-prime equation of state.
    This equation is derived from the assumption that K' = b*(V/V_0)^a.

    This equation of state has no temperature dependence.
    """

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        ai, bi, ci = make_params(
            params["K_0"], params["Kprime_0"], params["Kprime_inf"], params["Kdprime_0"]
        )

        lnVrel = np.log(volume / params["V_0"])
        return params["K_0"] * np.exp(-bi * (np.exp(ai * lnVrel) - 1.0) - ci * lnVrel)

    def volume(self, pressure, temperature, params):
        """
        Get the Vinet volume at a reference temperature for a given
        pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
        """

        def delta_pressure(x):
            return self.pressure(0.0, x, params) - pressure

        V = opt.brentq(delta_pressure, 0.1 * params["V_0"], 1.5 * params["V_0"])
        return V

    def pressure(self, temperature, volume, params):
        """
        Returns pressure :math:`[Pa]` as a function of volume :math:`[m^3]`.
        """
        ai, bi, ci = make_params(
            params["K_0"], params["Kprime_0"], params["Kprime_inf"], params["Kdprime_0"]
        )
        lnVrel = np.log(volume / params["V_0"])
        return params["P_0"] + (
            params["K_0"]
            * np.exp(bi)
            / ai
            * np.power(bi, ci / ai)
            * (generalised_gammainc(-ci / ai, bi * np.exp(ai * lnVrel), bi))
        )

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        ai, bi, ci = make_params(
            params["K_0"], params["Kprime_0"], params["Kprime_inf"], params["Kdprime_0"]
        )
        lnVrel = np.log(volume / params["V_0"])
        f = (
            -params["V_0"]
            * params["K_0"]
            * np.exp(bi)
            / ai
            * np.power(bi, (ci - 1.0) / ai)
        )

        Vrel = np.exp(lnVrel)
        I1 = (
            np.power(bi, 1.0 / ai)
            * Vrel
            * (generalised_gammainc(-ci / ai, bi * np.exp(ai * lnVrel), bi))
        )
        I2 = generalised_gammainc((1.0 - ci) / ai, bi * np.exp(ai * lnVrel), bi)

        return params["E_0"] + params["P_0"] * (volume - params["V_0"]) + f * (I1 - I2)

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + pressure * volume
        )

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 1.0e99

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return a very small number. :math:`[J/K/mol]`
        """
        return 1.0e-99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return a very small number. :math:`[J/K/mol]`
        """
        return 1.0e-99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters.
        The value for :math:`K'_{\\infty}` is thermodynamically bounded
        between 5/3 and :math:`K'_0` :cite:`StaceyDavis2004`.
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 1.0e5

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0", "Kdprime_0", "Kprime_inf"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["Kdprime_0"] > 0.0:
            warnings.warn("Kdprime_0 should be negative", stacklevel=2)
        if (
            -params["Kdprime_0"] * params["K_0"]
            < params["Kprime_0"] - params["Kprime_inf"]
        ):
            warnings.warn(
                "Kdprime_0*K_0 is expected to be more "
                "negative than (Kprime_0 - Kprime_inf)",
                stacklevel=2,
            )
        if (
            params["Kprime_inf"] < 5.0 / 3.0
            or params["Kprime_inf"] > params["Kprime_0"]
        ):
            warnings.warn(
                "Kprime_inf is expected to be greater "
                "than the Thomas-Fermi limit (5/3)",
                stacklevel=2,
            )



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts a tabulated version of the
# Stixrude and Lithgow-Bertelloni data format into the standard burnman
# format (printed to stdout)


import sys


def read_dataset(datafile):
    f = open(datafile, "r")
    ds = []
    for line in f:
        ds.append(line.decode("utf-8").split())
    return ds


ds = read_dataset("HHPH2013_endmembers.dat")
print(
    "# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences"
)
print(
    "# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU GPL v2 or later."
)
print("")
print("")
print('"""')
print("HHPH_2013")
print("Minerals from Holland et al 2013 and references therein")
print("The values in this document are all in S.I. units,")
print("unlike those in the original paper")
print("File autogenerated using HHPHdata_to_burnman.py")
print('"""')
print("")
print("from burnman.mineral import Mineral")
print("from burnman.solution import Solution")
print("from burnman.solutionmodel import *")
print("from burnman.tools.chemistry import dictionarize_formula, formula_mass")
print("")

print('"""')
print("ENDMEMBERS")
print('"""')
print("")

param_scales = [
    -1.0,
    -1.0,  # not nubmers, so we won't scale
    1.0e3,
    1.0e3,  # kJ -> J
    1.0,  # J/K/mol
    1.0e-5,  # kJ/kbar/mol -> m^3/mol
    1.0e3,
    1.0e-2,
    1.0e3,
    1.0e3,
    # kJ -> J and table conversion for b
    1.0e-5,  # table conversion
    1.0e8,  # kbar -> Pa
    1.0,  # no scale for K'0
    1.0e-8,
]  # GPa -> Pa # no scale for eta_s


formula = "0"
for idx, m in enumerate(ds):
    if idx == 0:
        param_names = m
    else:
        print("class", m[0].lower(), "(Mineral):")
        print("    def __init__(self):")
        print("".join(["        formula='", m[1], "'"]))
        print("        formula = dictionarize_formula(formula)")
        print("        self.params = {")
        print("".join(["            'name': '", m[0], "',"]))
        print("            'formula': formula,")
        print("            'equation_of_state': 'hp_tmt',")
        for pid, param in enumerate(m):
            if pid > 1 and pid != 3 and pid < 6:
                print(
                    "            '" + param_names[pid] + "':",
                    float(param) * param_scales[pid],
                    ",",
                )

        print(
            "            'Cp':",
            [round(float(m[i]) * param_scales[i], 10) for i in [6, 7, 8, 9]],
            ",",
        )
        for pid, param in enumerate(m):
            if pid > 9:
                print(
                    "            '" + param_names[pid] + "':",
                    float(param) * param_scales[pid],
                    ",",
                )

        print("            'n': sum(formula.values()),")
        print("            'molar_mass': formula_mass(formula)}")
        print("")
        print("        self.uncertainties = {")
        print(
            "            '" + param_names[3] + "':", float(m[3]) * param_scales[3], "}"
        )

        print("        Mineral.__init__(self)")
        print("")



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from . import equation_of_state as eos
from ..utils.math import bracket
import warnings
import numpy as np


def bulk_modulus(volume, params):
    """
    Compute the bulk modulus as per the Morse potential
    equation of state.
    Returns bulk modulus in the same units as
    the reference bulk modulus.
    Pressure must be in :math:`[Pa]`.
    """

    VoverV0 = volume / params["V_0"]
    x = (params["Kprime_0"] - 1.0) * (1.0 - np.power(VoverV0, 1.0 / 3.0))
    K = params["K_0"] * (
        (
            2.0
            / (params["Kprime_0"] - 1.0)
            * np.power(VoverV0, -2.0 / 3.0)
            * (np.exp(2.0 * x) - np.exp(x))
        )
        + (np.power(VoverV0, -1.0 / 3.0) * (2.0 * np.exp(2.0 * x) - np.exp(x)))
    )
    return K


def shear_modulus(volume, params):
    """
    Shear modulus not currently implemented for this equation of state
    """
    return 0.0


def morse_potential(VoverV0, params):
    """
    Equation for the Morse Potential equation of state,
    returns pressure in the same units that are supplied
    for the reference bulk modulus (params['K_0'])
    """
    x = (params["Kprime_0"] - 1.0) * (1.0 - np.power(VoverV0, 1.0 / 3.0))
    return (
        3.0
        * params["K_0"]
        / (params["Kprime_0"] - 1.0)
        * np.power(VoverV0, -2.0 / 3.0)
        * (np.exp(2.0 * x) - np.exp(x))
    ) + params["P_0"]


def volume(pressure, params):
    """
    Get the Morse Potential volume at a
    reference temperature for a given pressure :math:`[Pa]`.
    Returns molar volume in :math:`[m^3]`
    """
    func = lambda V: morse_potential(V / params["V_0"], params) - pressure
    try:
        sol = bracket(func, params["V_0"], 1.0e-2 * params["V_0"])
    except:
        raise ValueError(
            "Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?"
        )
    return opt.brentq(func, sol[0], sol[1])


class Morse(eos.EquationOfState):
    """
    Class for the isothermal Morse Potential equation of state
    detailed in :cite:`Stacey1981`.
    This equation of state has no temperature dependence.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return morse_potential(volume / params["V_0"], params)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return shear_modulus(volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """

        x = (params["Kprime_0"] - 1) * (1 - np.power(volume / params["V_0"], 1.0 / 3.0))
        intPdV = (
            9.0
            / 2.0
            * params["V_0"]
            * params["K_0"]
            / np.power(params["Kprime_0"] - 1.0, 2.0)
            * (2.0 * np.exp(x) - np.exp(2.0 * x) - 1.0)
        )

        return -intPdV + params["E_0"]

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + volume * pressure
        )

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 0.0

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if "G_0" not in params:
            params["G_0"] = float("nan")
        if "Gprime_0" not in params:
            params["Gprime_0"] = float("nan")

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0", "G_0", "Gprime_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["G_0"] < 0.0 or params["G_0"] > 1.0e13:
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0:
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Utils
-----

The utils submodule of BurnMan contains utility functions
that can be used by the other BurnMan modules.

They do not depend on any other BurnMan objects apart from
those in constants.py.
"""

from __future__ import absolute_import

from . import chemistry
from . import math
from . import misc
from . import reductions
from . import unitcell
from . import geotherm
from . import anisotropy



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts a tabulated version of the
# Stixrude and Lithgow-Bertelloni data format into the standard burnman
# format (printed to stdout)


import sys
import os.path

if os.path.isfile("stx11ver.dat") == False:
    print("This code requires the data file stx11ver.dat.")
    print("This file is bundled with the Perple_X data files, which can be found here:")
    print("http://www.perplex.ethz.ch/perplex/ibm_and_mac_archives/")
    print("")
    print("Please download the file and place it in this directory.")
    exit()


def read_dataset(datafile):
    f = open(datafile, "r")
    ds = []
    for line in f:
        ds.append(line.split())
    return ds


def read_dataset_utf8(datafile):
    f = open(datafile, "r")
    ds = []
    for line in f:
        ds.append(line.decode("utf-8").split())
    return ds


def process_stixrude():
    perplex_ds = read_dataset("stx11ver.dat")
    alias_mapping = read_dataset("perplex_slb_2011_names.txt")

    alias = {}
    accurate_values = {}
    configurational_entropies = {}
    for line in alias_mapping:
        alias[line[0]] = line[2]

    process = 0  # flag to process data or not (to cut out first lines)
    n = 0  # line number for each record
    for line in perplex_ds:
        if len(line) != 0:
            if process == 1:
                if line[0] == "end":
                    n = 0
                else:
                    n = n + 1
                if n == 1:
                    name = alias[line[0]]
                    data = []
                    configurational_entropy = "None"
                elif n == 3:
                    # F0, (n, ignored), -V0
                    data.extend([float(line[2]), -1.0 * float(line[8])])
                elif n == 4:
                    # K_0, K_prime, Debye_0, grueneisen_0, q_0, eta_s0,
                    # [configurational_and_magnetic_entropy]
                    data.extend(
                        [
                            float(line[2]),
                            float(line[5]),
                            float(line[8]),
                            float(line[11]),
                            float(line[14]),
                            float(line[17]),
                        ]
                    )
                    if len(line) == 21:
                        configurational_entropy = float(line[20])
                elif n == 5:
                    # mu_S0, mu_S0_prime
                    data.extend([float(line[2]), float(line[5])])
                    # Sort into correct order for processing: F, V, K, K',
                    #                        Debye, gruen, q, G, Gprime, etaS0
                    accurate_values[name] = [
                        data[0] / 1.0e3,
                        data[1] * 10.0,
                        data[2] / 1.0e4,
                        data[3],
                        data[4],
                        data[5],
                        data[6],
                        data[8] / 1.0e4,
                        data[9],
                        data[7],
                    ]
                    configurational_entropies[name] = configurational_entropy

            if line[0] == "end_components":
                process = 1

    return accurate_values, configurational_entropies


accurate_values, configurational_entropies = process_stixrude()


ds = read_dataset_utf8("slb_2011.txt")
landau = read_dataset_utf8("slb_2011_landau.txt")
landau_params = {}
for line in landau:
    phase = str(line[0]).lower()
    if phase != "#":
        landau_params[phase] = [float(line[1]), float(line[2]) * 1.0e-6, float(line[3])]

print(
    "# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences"
)
print(
    "# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU GPL v2 or later."
)
print("")
print("")
print('"""')
print("SLB_2011")
print("Minerals from Stixrude & Lithgow-Bertelloni 2011 and references therein")
print("File autogenerated using SLBdata_to_burnman.py")
print('"""')
print("")
print("from __future__ import absolute_import")
print("")
print("from ..mineral import Mineral")
print("from ..solution import Solution")
print("from ..solutionmodel import *")
print("from ..tools.chemistry import dictionarize_formula, formula_mass")
print("")

param_scales = [
    -1.0,
    -1.0,  # not numbers, so we won't scale
    1.0e3,
    1.0e3,  # KJ -> J
    1.0e-6,
    1.0e-6,  # cm^3/mol -> m^3/mol
    1.0e9,
    1.0e9,  # GPa -> Pa
    1.0,
    1.0,  # no scale for K'
    1.0,
    1.0,  # no scale for Debye
    1.0,
    1.0,  # no scale for gruneisen
    1.0,
    1.0,  # no scale for q
    1.0e9,
    1.0e9,  # GPa -> Pa
    1.0,
    1.0,  # no scale for G'
    1.0,
    1.0,
]  # no scale for eta_s


solutionfile = "slb_2011_solutions.txt"
with open(solutionfile, "r") as fin:
    print(fin.read())
fin.close()

print('"""')
print("ENDMEMBERS")
print('"""')
print("")
formula = "0"
for idx, m in enumerate(ds):
    if idx == 0:
        param_names = m
    else:
        print("class", m[0].lower(), "(Mineral):")
        print("    def __init__(self):")
        print("".join(["        formula='", m[1], "'"]))
        print("        formula = dictionarize_formula(formula)")
        print("        self.params = {")
        print("".join(["            'name': '", m[0], "',"]))
        print("            'formula': formula,")
        print("            'equation_of_state': 'slb3',")
        for pid, param in enumerate(accurate_values[m[0].lower()]):
            print(
                "            '" + param_names[(pid + 1) * 2] + "':",
                float(param) * param_scales[(pid + 1) * 2],
                ",",
            )
        print("            'n': sum(formula.values()),")
        print("            'molar_mass': formula_mass(formula)}")
        print("")
        if (
            landau_params.has_key(m[0].lower())
            or configurational_entropies[m[0].lower()] != "None"
        ):
            print(
                "        self.property_modifiers = [",
            )
            if landau_params.has_key(m[0].lower()):
                print(
                    "['landau', {'Tc_0':",
                    landau_params[m[0].lower()][0],
                    ", 'S_D':",
                    landau_params[m[0].lower()][2],
                    ", 'V_D':",
                    landau_params[m[0].lower()][1],
                    "}]",
                )
            if configurational_entropies[m[0].lower()] != "None":
                print(
                    "['linear', {'delta_E':",
                    0.0,
                    ", 'delta_S':",
                    configurational_entropies[m[0].lower()],
                    ", 'delta_V':",
                    0.0,
                    "}]",
                )
            print("]")
            print("")
        print("        self.uncertainties = {")
        for pid, param in enumerate(m):
            if pid > 1 and pid % 2 == 1 and pid < 21:
                print(
                    "            '" + param_names[pid] + "':",
                    float(param) * param_scales[pid],
                    ",",
                )
        pid = 21
        param = m[pid]
        print(
            "            '" + param_names[pid] + "':",
            float(param) * param_scales[pid],
            "}",
        )
        print("        Mineral.__init__(self)")
        print("")


aliasfile = "slb_2011_aliases.txt"
with open(aliasfile, "r") as fin:
    print(fin.read())
fin.close()



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
SLB_2011_ZSB_2013
^^^^^^^^^^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2011, Zhang, Stixrude & Brodholt 2013, and references therein.

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class stishovite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 14.02e-6,
            "K_0": 314.0e9,
            "Kprime_0": 3.8,
            "G_0": 220.0e9,
            "Gprime_0": 1.9,
            "molar_mass": 0.0601,
            "n": 3,
            "Debye_0": 1108.0,
            "grueneisen_0": 1.37,
            "q_0": 2.8,
            "eta_s_0": 4.6,
        }

        self.uncertainties = {
            "err_K_0": 8.0e9,
            "err_Kprime_0": 0.1,
            "err_G_0": 12.0e9,
            "err_Gprime_0": 0.1,
            "err_Debye_0": 13.0,
            "err_grueneisen_0": 0.17,
            "err_q_0": 2.2,
            "err_eta_s_0": 1.0,
        }

        Mineral.__init__(self)


class periclase(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 11.24e-6,
            "K_0": 161.0e9,
            "Kprime_0": 3.8,
            "G_0": 131.0e9,
            "Gprime_0": 2.1,
            "molar_mass": 0.0403,
            "n": 2,
            "Debye_0": 767.0,
            "grueneisen_0": 1.36,
            "q_0": 1.7,  # 1.7
            "eta_s_0": 2.8,
        }  # 2.8

        self.uncertainties = {
            "err_K_0": 3.0e9,
            "err_Kprime_0": 0.2,
            "err_G_0": 1.0e9,
            "err_Gprime_0": 0.1,
            "err_Debye_0": 9.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.2,
            "err_eta_s_0": 0.2,
        }

        Mineral.__init__(self)


class wuestite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 12.26e-6,
            "K_0": 179.0e9,
            "Kprime_0": 4.9,
            "G_0": 59.0e9,
            "Gprime_0": 1.4,
            "molar_mass": 0.0718,
            "n": 2,
            "Debye_0": 454.0,
            "grueneisen_0": 1.53,
            "q_0": 1.7,  # 1.7
            "eta_s_0": -0.1,
        }

        self.uncertainties = {
            "err_K_0": 1.0e9,
            "err_Kprime_0": 0.2,
            "err_G_0": 1.0e9,
            "err_Gprime_0": 0.1,
            "err_Debye_0": 21.0,
            "err_grueneisen_0": 0.13,
            "err_q_0": 1.0,
            "err_eta_s_0": 1.0,
        }

        Mineral.__init__(self)


class mg_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 24.45e-6,
            "K_0": 250.5e9,
            "Kprime_0": 4.01,
            "G_0": 172.9e9,
            "Gprime_0": 1.74,
            "molar_mass": 0.1000,
            "n": 5,
            "Debye_0": 905.9,
            "grueneisen_0": 1.44,
            "q_0": 1.09,
            "eta_s_0": 2.13,
        }  # 2.6

        self.uncertainties = {
            "err_K_0": 3.0e9,
            "err_Kprime_0": 0.1,
            "err_G_0": 2.0e9,
            "err_Gprime_0": 0.0,
            "err_Debye_0": 5.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.3,
            "err_eta_s_0": 0.3,
        }

        Mineral.__init__(self)


class fe_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 25.49e-6,
            "K_0": 272.0e9,
            "Kprime_0": 4.1,
            "G_0": 133.0e9,
            "Gprime_0": 1.4,
            "molar_mass": 0.1319,
            "n": 5,
            "Debye_0": 871.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "eta_s_0": 2.3,
        }  # 2.3

        self.uncertainties = {
            "err_K_0": 40e9,
            "err_Kprime_0": 1.0,
            "err_G_0": 40e9,
            "err_Gprime_0": 0.0,
            "err_Debye_0": 26.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_eta_s_0": 1.0,
        }

        Mineral.__init__(self)


mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.classes.calibrant import Calibrant
import numpy as np


"""
Dubrovinsky_1998
^^^^^^^^^^^^^^^^
"""


class Al2O3_corundum(Calibrant):
    """
    The Al2O3 pressure standard reported by
    Dubrovinsky (1998; https://dx.doi.org/10.1007/s002690050133).
    """

    def __init__(self):
        def _pressure_Dubrovinsky_Al2O3(volume, temperature, params):

            # Modified BM3+Thermal
            a0 = params["a"]
            a1 = params["b"]
            a2 = params["c"]
            K0 = params["K_0"]
            Kprime0 = params["Kprime_0"]
            dK_dT = params["dK_dT_P"]
            dKprime_dT = params["dKprime_dT"]

            KT = (K0) + dK_dT * (temperature - params["T_0"])
            KprimeT = Kprime0 + dKprime_dT * (temperature - params["T_0"])
            V0T = params["V_0"] * np.exp(
                (
                    (a0 * temperature)
                    + (a1 * ((temperature**2.0) / 2.0))
                    - (a2 / temperature)
                )
                - (
                    (a0 * params["T_0"])
                    + (a1 * ((params["T_0"] ** 2.0) / 2.0))
                    - (a2 / params["T_0"])
                )
            )
            V_V0T = volume / V0T
            f = 0.5 * ((V_V0T ** (-2.0 / 3.0)) - 1.0)

            P = (3.0 * KT * f * (1.0 + 2.0 * f) ** (5.0 / 2.0)) * (
                1.0 - 3.0 / 2.0 * (4.0 - KprimeT) * f
            )

            return P

        _params_Dubrovinsky_Al2O3 = {
            "V_0": 2.558e-05,
            "K_0": 253.0e9,
            "Kprime_0": 5.0,
            "a": 2.515e-5,
            "b": 1.958e-9,
            "c": -0.305,
            "dK_dT_P": -0.02e9,
            "dKprime_dT": 0.0,
            "n": 5.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 6.0,
        }

        Calibrant.__init__(
            self, _pressure_Dubrovinsky_Al2O3, "pressure", _params_Dubrovinsky_Al2O3
        )



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Tools
-----

The tools submodule of BurnMan is designed to provide high level
functions that use other BurnMan submodules.
It does not contain any functions that are required by the core
BurnMan modules.
"""

from __future__ import absolute_import

from . import chemistry
from . import eos
from . import equilibration
from . import output_seismo
from . import partitioning
from . import plot
from . import polytope
from . import solution
from . import unitcell
